# standard library imports
from collections import OrderedDict
import ctypes
from enum import IntEnum
import os
import queue
import re
import struct
import warnings

# 3rd party library imports
import numpy as np

# Local imports
from ..config import glymur_config

# The error messages queue
EQ = queue.Queue()

loader = ctypes.windll.LoadLibrary if os.name == "nt" else ctypes.CDLL
_LIBTIFF = glymur_config("tiff")
_LIBC = glymur_config("c")


class LibTIFFError(RuntimeError):
    """Raise this exception if we detect a generic error from libtiff."""

    pass


class Compression(IntEnum):
    """Compression scheme used on the image data.

    See Also
    --------
    Photometric : The color space of the image data.
    """

    NONE = 1
    CCITTRLE = 2  # CCITT modified Huffman RLE
    CCITTFAX3 = 3  # CCITT Group 3 fax encoding
    CCITT_T4 = 3  # CCITT T.4 (TIFF 6 name)
    CCITTFAX4 = 4  # CCITT Group 4 fax encoding
    CCITT_T6 = 4  # CCITT T.6 (TIFF 6 name)
    LZW = 5  # Lempel-Ziv  & Welch
    OJPEG = 6  # 6.0 JPEG
    JPEG = 7  # JPEG DCT compression
    T85 = 9  # TIFF/FX T.85 JBIG compression
    T43 = 10  # TIFF/FX T.43 colour by layered JBIG compression
    NEXT = 32766  # NeXT 2-bit RLE
    CCITTRLEW = 32771  # #1 w/ word alignment
    PACKBITS = 32773  # Macintosh RLE
    THUNDERSCAN = 32809  # ThunderScan RLE
    PIXARFILM = 32908  # companded 10bit LZW
    PIXARLOG = 32909  # companded 11bit ZIP
    DEFLATE = 32946  # compression
    ADOBE_DEFLATE = 8  # compression, as recognized by Adobe
    DCS = 32947  # DCS encoding
    JBIG = 34661  # JBIG
    SGILOG = 34676  # Log Luminance RLE
    SGILOG24 = 34677  # Log 24-bit packed
    JP2000 = 34712  # JPEG2000
    LZMA = 34925  # LZMA2


class InkSet(IntEnum):
    """
    The set of inks used in a separated (PhotometricInterpretation=5) image.
    """

    CMYK = 1
    MULTIINK = 2


class JPEGColorMode(IntEnum):
    """When writing images with photometric interpretation equal to YCbCr and
    compression equal to JPEG, the pseudo tag JPEGColorMode should usually be
    set to RGB, unless the image values truly are in YCbCr.

    See Also
    --------
    Photometric : The color space of the image data.
    """

    RAW = 0
    RGB = 1


class PlanarConfig(IntEnum):
    """How the components of each pixel are stored.

    Writing images with a PlanarConfig value of PlanarConfig.SEPARATE is not
    currently supported.
    """

    CONTIG = 1  # single image plane
    SEPARATE = 2  # separate planes of data


class Orientation(IntEnum):
    """The orientation of the image with respect to the rows and columns."""

    TOPLEFT = 1  # row 0 top, col 0 lhs */
    TOPRIGHT = 2  # row 0 top, col 0 rhs */
    BOTRIGHT = 3  # row 0 bottom, col 0 rhs */
    BOTLEFT = 4  # row 0 bottom, col 0 lhs */
    LEFTTOP = 5  # row 0 lhs, col 0 top */
    RIGHTTOP = 6  # row 0 rhs, col 0 top */
    RIGHTBOT = 7  # row 0 rhs, col 0 bottom */
    LEFTBOT = 8  # row 0 lhs, col 0 bottom */


class Photometric(IntEnum):
    """The color space of the image data.

    Examples
    --------

    Load an image of astronaut Eileen Collins from scikit-image.

    >>> import numpy as np
    >>> import skimage.data
    >>> image = skimage.data.astronaut()

    Create a BigTIFF with JPEG compression.  There is not much reason to do
    this if you do not also specify YCbCr as the photometric interpretation.

    >>> w, h, nz = image.shape
    >>> tw, th = w // 2, h // 2
    >>> from glymur.lib import tiff as libtiff
    >>> fp = libtiff.open('astronaut-jpeg.tif', mode='w8')
    >>> libtiff.setField(fp, 'Photometric', libtiff.Photometric.YCBCR)
    >>> libtiff.setField(fp, 'Compression', libtiff.Compression.JPEG)
    >>> libtiff.setField(fp, 'JPEGColorMode', libtiff.JPEGColorMode.RGB)
    >>> libtiff.setField(fp, 'PlanarConfig', libtiff.PlanarConfig.CONTIG)
    >>> libtiff.setField(fp, 'JPEGQuality', 90)
    >>> libtiff.setField(fp, 'YCbCrSubsampling', 1, 1)
    >>> libtiff.setField(fp, 'ImageWidth', w)
    >>> libtiff.setField(fp, 'ImageLength', h)
    >>> libtiff.setField(fp, 'TileWidth', tw)
    >>> libtiff.setField(fp, 'TileLength', th)
    >>> libtiff.setField(fp, 'BitsPerSample', 8)
    >>> libtiff.setField(fp, 'SamplesPerPixel', nz)
    >>> libtiff.setField(fp, 'Software', libtiff.getVersion())
    >>> libtiff.writeEncodedTile(fp, 0, image[:th, :tw].copy())
    >>> libtiff.writeEncodedTile(fp, 1, image[:th, tw:w].copy())
    >>> libtiff.writeEncodedTile(fp, 2, image[th:h, :tw].copy())
    >>> libtiff.writeEncodedTile(fp, 3, image[th:h, tw:w].copy())
    >>> libtiff.close(fp)
    """

    MINISWHITE = 0  # value is white
    MINISBLACK = 1  # value is black
    RGB = 2  # color model
    PALETTE = 3  # map indexed
    MASK = 4  # holdout mask
    SEPARATED = 5  # color separations
    YCBCR = 6  # CCIR 601
    CIELAB = 8  # 1976 CIE L*a*b*
    ICCLAB = 9  # L*a*b* [Adobe TIFF Technote 4]
    ITULAB = 10  # L*a*b*
    CFA = 32803  # filter array
    LOGL = 32844  # Log2(L)
    LOGLUV = 32845  # Log2(L) (u',v')


class SampleFormat(IntEnum):
    """Specifies how to interpret each data sample in a pixel."""

    UINT = 1
    INT = 2
    IEEEFP = 3
    VOID = 4
    COMPLEXINT = 5
    COMPLEXIEEEP = 6


def _handle_error(module, fmt, ap):
    # Use VSPRINTF in the C library to put together the error message.
    # int vsprintf(char * buffer, const char * restrict format, va_list ap);
    buffer = ctypes.create_string_buffer(1000)

    argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p]
    _LIBC.vsprintf.argtypes = argtypes
    _LIBC.vsprintf.restype = ctypes.c_int32
    _LIBC.vsprintf(buffer, fmt, ap)

    module = module.decode("utf-8")
    error_str = buffer.value.decode("utf-8")

    message = f"{module}: {error_str}"
    EQ.put(message)
    return None


def _handle_warning(module, fmt, ap):
    # Use VSPRINTF in the C library to put together the warning message.
    # int vsprintf(char * buffer, const char * restrict format, va_list ap);
    buffer = ctypes.create_string_buffer(1000)

    argtypes = [ctypes.c_char_p, ctypes.c_char_p, ctypes.c_void_p]
    _LIBC.vsprintf.argtypes = argtypes
    _LIBC.vsprintf.restype = ctypes.c_int32
    _LIBC.vsprintf(buffer, fmt, ap)

    module = module.decode("utf-8")
    warning_str = buffer.value.decode("utf-8")

    message = f"{module}: {warning_str}"
    warnings.warn(message)


# Set the function types for the warning handler.
_WFUNCTYPE = ctypes.CFUNCTYPE(
    ctypes.c_void_p,  # return type of warning handler, void *
    ctypes.c_char_p,  # module
    ctypes.c_char_p,  # fmt
    ctypes.c_void_p,  # va_list
)

_ERROR_HANDLER = _WFUNCTYPE(_handle_error)
_WARNING_HANDLER = _WFUNCTYPE(_handle_warning)


def _set_error_warning_handlers():
    """Setup default python error and warning handlers."""
    old_warning_handler = setWarningHandler()
    old_error_handler = setErrorHandler()

    return old_error_handler, old_warning_handler


def _reset_error_warning_handlers(old_error_handler, old_warning_handler):
    """Restore previous error and warning handlers."""
    setWarningHandler(old_warning_handler)
    setErrorHandler(old_error_handler)


def close(fp):
    """Corresponds to TIFFClose"""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [ctypes.c_void_p]
    _LIBTIFF.TIFFClose.argtypes = ARGTYPES
    _LIBTIFF.TIFFClose.restype = None
    _LIBTIFF.TIFFClose(fp)

    _reset_error_warning_handlers(err_handler, warn_handler)


def computeStrip(fp, row, sample):
    """Corresponds to TIFFComputeStrip"""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [ctypes.c_void_p, ctypes.c_uint32, ctypes.c_uint16]
    _LIBTIFF.TIFFComputeStrip.argtypes = ARGTYPES
    _LIBTIFF.TIFFComputeStrip.restype = ctypes.c_uint32
    stripnum = _LIBTIFF.TIFFComputeStrip(fp, row, sample)

    _reset_error_warning_handlers(err_handler, warn_handler)

    return stripnum


def computeTile(fp, x, y, z, sample):
    """Corresponds to TIFFComputeTile"""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [
        ctypes.c_void_p,
        ctypes.c_uint32,
        ctypes.c_uint32,
        ctypes.c_uint32,
        ctypes.c_uint16,
    ]
    _LIBTIFF.TIFFComputeTile.argtypes = ARGTYPES
    _LIBTIFF.TIFFComputeTile.restype = ctypes.c_uint32
    tilenum = _LIBTIFF.TIFFComputeTile(fp, x, y, z, sample)

    _reset_error_warning_handlers(err_handler, warn_handler)

    return tilenum


def isTiled(fp):
    """Corresponds to TIFFIsTiled"""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [ctypes.c_void_p]

    _LIBTIFF.TIFFIsTiled.argtypes = ARGTYPES
    _LIBTIFF.TIFFIsTiled.restype = ctypes.c_int

    status = _LIBTIFF.TIFFIsTiled(fp)

    _reset_error_warning_handlers(err_handler, warn_handler)

    return status


def numberOfStrips(fp):
    """Corresponds to TIFFNumberOfStrips."""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [ctypes.c_void_p]
    _LIBTIFF.TIFFNumberOfStrips.argtypes = ARGTYPES
    _LIBTIFF.TIFFNumberOfStrips.restype = ctypes.c_uint32

    numstrips = _LIBTIFF.TIFFNumberOfStrips(fp)

    _reset_error_warning_handlers(err_handler, warn_handler)

    return numstrips


def numberOfTiles(fp):
    """Corresponds to TIFFNumberOfTiles."""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [ctypes.c_void_p]
    _LIBTIFF.TIFFNumberOfTiles.argtypes = ARGTYPES
    _LIBTIFF.TIFFNumberOfTiles.restype = ctypes.c_uint32

    numtiles = _LIBTIFF.TIFFNumberOfTiles(fp)

    _reset_error_warning_handlers(err_handler, warn_handler)

    return numtiles


def readEncodedStrip(fp, stripnum, strip, size=-1):
    """Corresponds to TIFFReadEncodedStrip"""
    err_handler, warn_handler = _set_error_warning_handlers()

    if size == -1:
        size = strip.nbytes

    ARGTYPES = [
        ctypes.c_void_p,
        ctypes.c_uint32,
        ctypes.c_void_p,
        ctypes.c_int32
    ]
    _LIBTIFF.TIFFReadEncodedStrip.argtypes = ARGTYPES
    _LIBTIFF.TIFFReadEncodedStrip.restype = check_error
    _LIBTIFF.TIFFReadEncodedStrip(
        fp, stripnum, strip.ctypes.data_as(ctypes.c_void_p), size
    )

    _reset_error_warning_handlers(err_handler, warn_handler)

    return strip


def readEncodedTile(fp, tilenum, tile, size=-1):
    """Corresponds to TIFFComputeTile"""
    err_handler, warn_handler = _set_error_warning_handlers()

    if size == -1:
        size = tile.nbytes

    ARGTYPES = [
        ctypes.c_void_p,
        ctypes.c_uint32,
        ctypes.c_void_p,
        ctypes.c_int32
    ]
    _LIBTIFF.TIFFReadEncodedTile.argtypes = ARGTYPES
    _LIBTIFF.TIFFReadEncodedTile.restype = check_error
    _LIBTIFF.TIFFReadEncodedTile(
        fp,
        tilenum,
        tile.ctypes.data_as(ctypes.c_void_p),
        -1
    )

    _reset_error_warning_handlers(err_handler, warn_handler)

    return tile


def readRGBAStrip(fp, row, strip):
    """Corresponds to TIFFReadRGBAStrip"""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [ctypes.c_void_p, ctypes.c_uint32, ctypes.c_void_p]
    _LIBTIFF.TIFFReadRGBAStrip.argtypes = ARGTYPES
    _LIBTIFF.TIFFReadRGBAStrip.restype = check_error
    _LIBTIFF.TIFFReadRGBAStrip(fp, row, strip.ctypes.data_as(ctypes.c_void_p))

    _reset_error_warning_handlers(err_handler, warn_handler)

    return strip


def readRGBATile(fp, x, y, tile):
    """Corresponds to TIFFReadRGBATile"""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [
        ctypes.c_void_p,
        ctypes.c_uint32,
        ctypes.c_uint32,
        ctypes.c_void_p
    ]
    _LIBTIFF.TIFFReadRGBATile.argtypes = ARGTYPES
    _LIBTIFF.TIFFReadRGBATile.restype = check_error
    _LIBTIFF.TIFFReadRGBATile(fp, x, y, tile.ctypes.data_as(ctypes.c_void_p))

    _reset_error_warning_handlers(err_handler, warn_handler)

    return tile


def readRGBAImageOriented(
    fp,
    width=None,
    height=None,
    orientation=Orientation.TOPLEFT
):
    """Read an image as if it were RGBA.

    This function corresponds to the TIFFReadRGBAImageOriented function in the
    libtiff library.

    Parameters
    ----------
    fp : ctypes void pointer
        File pointer returned by libtiff.
    width, height : int
        Width and height of the returned image.
    orientation : int
        The raster origin position.

    See Also
    --------
    Orientation
    """
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [
        ctypes.c_void_p,
        ctypes.c_uint32,
        ctypes.c_uint32,
        ctypes.POINTER(ctypes.c_uint32),
        ctypes.c_int32,
        ctypes.c_int32,
    ]

    _LIBTIFF.TIFFReadRGBAImageOriented.argtypes = ARGTYPES
    _LIBTIFF.TIFFReadRGBAImageOriented.restype = check_error

    if width is None:
        width = getFieldDefaulted(fp, "ImageWidth")
    if height is None:
        height = getFieldDefaulted(fp, "ImageLength")

    img = np.zeros((height, width, 4), dtype=np.uint8)
    raster = img.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32))
    _LIBTIFF.TIFFReadRGBAImageOriented(
        fp,
        width,
        height,
        raster,
        orientation,
        0
    )

    _reset_error_warning_handlers(err_handler, warn_handler)

    return img


def writeEncodedStrip(fp, stripnum, stripdata, size=-1):
    """Corresponds to TIFFWriteEncodedStrip."""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [
        ctypes.c_void_p,
        ctypes.c_uint32,
        ctypes.c_void_p,
        ctypes.c_uint32
    ]
    _LIBTIFF.TIFFWriteEncodedStrip.argtypes = ARGTYPES
    _LIBTIFF.TIFFWriteEncodedStrip.restype = check_error
    raster = stripdata.ctypes.data_as(ctypes.c_void_p)

    if size == -1:
        size = stripdata.nbytes

    _LIBTIFF.TIFFWriteEncodedStrip(fp, stripnum, raster, size)

    _reset_error_warning_handlers(err_handler, warn_handler)


def writeEncodedTile(fp, tilenum, tiledata, size=-1):
    """Corresponds to TIFFWriteEncodedTile."""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [
        ctypes.c_void_p,
        ctypes.c_uint32,
        ctypes.c_void_p,
        ctypes.c_uint32
    ]
    _LIBTIFF.TIFFWriteEncodedTile.argtypes = ARGTYPES
    _LIBTIFF.TIFFWriteEncodedTile.restype = check_error
    raster = tiledata.ctypes.data_as(ctypes.c_void_p)

    if size == -1:
        size = tiledata.nbytes

    _LIBTIFF.TIFFWriteEncodedTile(fp, tilenum, raster, size)

    _reset_error_warning_handlers(err_handler, warn_handler)


def RGBAImageOK(fp):
    """Corresponds to TIFFRGBAImageOK."""
    err_handler, warn_handler = _set_error_warning_handlers()

    emsg = ctypes.create_string_buffer(1024)
    ARGTYPES = [ctypes.c_void_p, ctypes.c_char_p]
    _LIBTIFF.TIFFRGBAImageOK.argtypes = ARGTYPES
    _LIBTIFF.TIFFRGBAImageOK.restype = ctypes.c_int
    ok = _LIBTIFF.TIFFRGBAImageOK(fp, emsg)

    _reset_error_warning_handlers(err_handler, warn_handler)

    if ok:
        return True
    else:
        return False


def getFieldDefaulted(fp, tag):
    """Corresponds to the TIFFGetFieldDefaulted library routine."""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [ctypes.c_void_p, ctypes.c_int32]

    tag_num = TAGS[tag]["number"]

    # Append the proper return type for the tag.
    tag_type = TAGS[tag]["type"]
    ARGTYPES.append(ctypes.POINTER(TAGS[tag]["type"]))
    _LIBTIFF.TIFFGetFieldDefaulted.argtypes = ARGTYPES

    _LIBTIFF.TIFFGetFieldDefaulted.restype = check_error

    # instantiate the tag value
    item = tag_type()
    _LIBTIFF.TIFFGetFieldDefaulted(fp, tag_num, ctypes.byref(item))

    _reset_error_warning_handlers(err_handler, warn_handler)

    return item.value


def getVersion():
    """Corresponds to the TIFFGetVersion library routine."""
    try:
        _LIBTIFF.TIFFGetVersion.restype = ctypes.c_char_p
    except AttributeError:
        # libtiff not installed
        return "0.0.0"

    v = _LIBTIFF.TIFFGetVersion().decode("utf-8")

    # v would be something like
    #
    # LIBTIFF, Version 4.3.0
    # Copyright (c) 1988-1996 Sam Leffler
    # Copyright (c) 1991-1996 Silicon Graphics, Inc.
    #
    # All we want is the '4.3.0'
    m = re.search(r"(?P<version>\d+\.\d+\.\d+)", v)
    return m.group("version")


def open(filename, mode="r"):
    """Corresponds to TIFFOpen

    Parameters
    ----------
    filename : path or str
        Path to TIFF
    """
    err_handler, warn_handler = _set_error_warning_handlers()

    filename = str(filename)

    ARGTYPES = [ctypes.c_char_p, ctypes.c_char_p]
    _LIBTIFF.TIFFOpen.argtypes = ARGTYPES
    _LIBTIFF.TIFFOpen.restype = ctypes.c_void_p
    file_argument = ctypes.c_char_p(filename.encode())
    mode_argument = ctypes.c_char_p(mode.encode())
    fp = _LIBTIFF.TIFFOpen(file_argument, mode_argument)
    check_error(fp)

    _reset_error_warning_handlers(err_handler, warn_handler)

    return fp


def setErrorHandler(func=_ERROR_HANDLER):
    # The signature of the error handler is
    #     const char *module, const char *fmt, va_list ap
    #
    # The return type is void *
    _LIBTIFF.TIFFSetErrorHandler.argtypes = [_WFUNCTYPE]
    _LIBTIFF.TIFFSetErrorHandler.restype = _WFUNCTYPE
    old_error_handler = _LIBTIFF.TIFFSetErrorHandler(func)
    return old_error_handler


def setField(fp, tag, *value):
    """Corresponds to TIFFSetField"""
    err_handler, warn_handler = _set_error_warning_handlers()

    ARGTYPES = [ctypes.c_void_p, ctypes.c_int32]

    # Append the proper return type for the tag.
    tag_num = TAGS[tag]["number"]
    tag_type = TAGS[tag]["type"]

    try:
        for ttype in tag_type:
            ARGTYPES.append(ttype)
    except TypeError:
        ARGTYPES.append(tag_type)

    _LIBTIFF.TIFFSetField.argtypes = ARGTYPES
    _LIBTIFF.TIFFSetField.restype = check_error

    if len(value) == 1 and tag_type == ctypes.c_char_p:
        _LIBTIFF.TIFFSetField(fp, tag_num, ctypes.c_char_p(value[0].encode()))
    elif len(value) == 1:
        _LIBTIFF.TIFFSetField(fp, tag_num, value[0])
    else:
        _LIBTIFF.TIFFSetField(fp, tag_num, *value)

    _reset_error_warning_handlers(err_handler, warn_handler)


def setWarningHandler(func=_WARNING_HANDLER):
    # The signature of the warning handler is
    #     const char *module, const char *fmt, va_list ap
    #
    # The return type is void *
    _LIBTIFF.TIFFSetWarningHandler.argtypes = [_WFUNCTYPE]
    _LIBTIFF.TIFFSetWarningHandler.restype = _WFUNCTYPE
    old_warning_handler = _LIBTIFF.TIFFSetWarningHandler(func)
    return old_warning_handler


def check_error(status):
    """Set a generic function as the restype attribute of all TIFF
    functions that return a int value.  This way we do not have to check
    for error status in each wrapping function and an exception will always be
    appropriately raised.
    """
    msg = ""
    while not EQ.empty():
        msg = EQ.get()
        raise LibTIFFError(msg)

    if status == 0:
        raise RuntimeError("failed")


TAGS = {
    "GPSVersionID": {
        "number": 0,
        "type": ctypes.c_char_p,
    },
    "GPSLatitudeRef": {
        "number": 1,
        "type": ctypes.c_char_p,
    },
    "GPSLatitude": {
        "number": 2,
        "type": ctypes.c_char_p,
    },
    "GPSLongitudeRef": {
        "number": 3,
        "type": ctypes.c_char_p,
    },
    "GPSLongitude": {
        "number": 4,
        "type": ctypes.c_char_p,
    },
    "GPSAltitudeRef": {
        "number": 5,
        "type": ctypes.c_char_p,
    },
    "GPSAltitude": {
        "number": 6,
        "type": ctypes.c_char_p,
    },
    "GPSTimestamp": {
        "number": 7,
        "type": ctypes.c_char_p,
    },
    "GPSSatellites": {
        "number": 8,
        "type": ctypes.c_char_p,
    },
    "GPSStatus": {
        "number": 9,
        "type": ctypes.c_char_p,
    },
    "GPSMeasureMode": {
        "number": 10,
        "type": ctypes.c_char_p,
    },
    "GPSDOP": {
        "number": 11,
        "type": ctypes.c_char_p,
    },
    "GPSSpeedRef": {
        "number": 12,
        "type": ctypes.c_char_p,
    },
    "GPSSpeed": {
        "number": 13,
        "type": ctypes.c_char_p,
    },
    "GPSTrackRef": {
        "number": 14,
        "type": ctypes.c_char_p,
    },
    "GPSTrack": {
        "number": 15,
        "type": ctypes.c_char_p,
    },
    "GPSImgDirectionRef": {
        "number": 16,
        "type": ctypes.c_char_p,
    },
    "GPSImgDirection": {
        "number": 17,
        "type": ctypes.c_char_p,
    },
    "GPSMapDatum": {
        "number": 18,
        "type": ctypes.c_char_p,
    },
    "GPSDestLatitudeRef": {
        "number": 19,
        "type": ctypes.c_char_p,
    },
    "GPSDestLatitude": {
        "number": 20,
        "type": ctypes.c_char_p,
    },
    "GPSDestLongitudeRef": {
        "number": 21,
        "type": ctypes.c_char_p,
    },
    "GPSDestLongitude": {
        "number": 22,
        "type": ctypes.c_char_p,
    },
    "GPSBearingRef": {
        "number": 23,
        "type": ctypes.c_char_p,
    },
    "GPSBearing": {
        "number": 24,
        "type": ctypes.c_char_p,
    },
    "GPSDestDistanceRef": {
        "number": 25,
        "type": ctypes.c_char_p,
    },
    "GPSDestDistance": {
        "number": 26,
        "type": ctypes.c_int16,
    },
    "GPSProcessingMethod": {
        "number": 27,
        "type": ctypes.c_char_p,
    },
    "GPSAreaInformation": {
        "number": 28,
        "type": ctypes.c_char_p,
    },
    "GPSDateStamp": {
        "number": 29,
        "type": ctypes.c_char_p,
    },
    "GPSDifferential": {
        "number": 30,
        "type": ctypes.c_char_p,
    },
    "GPSHPositioningError": {
        "number": 31,
        "type": ctypes.c_int16,
    },
    "SubFileType": {
        "number": 254,
        "type": ctypes.c_int16,
    },
    "OSubFileType": {
        "number": 255,
        "type": ctypes.c_int16,
    },
    "ImageWidth": {
        "number": 256,
        "type": ctypes.c_int32,
    },
    "ImageLength": {
        "number": 257,
        "type": ctypes.c_int32,
    },
    "BitsPerSample": {
        "number": 258,
        "type": ctypes.c_int16,
    },
    "Compression": {
        "number": 259,
        "type": ctypes.c_int16,
    },
    "Photometric": {
        "number": 262,
        "type": ctypes.c_int16,
    },
    "Threshholding": {
        "number": 263,
        "type": ctypes.c_int16,
    },
    "CellWidth": {
        "number": 264,
        "type": ctypes.c_int16,
    },
    "CellLength": {
        "number": 265,
        "type": ctypes.c_int16,
    },
    "FillOrder": {
        "number": 266,
        "type": ctypes.c_int16,
    },
    "DocumentName": {
        "number": 269,
        "type": ctypes.c_char_p,
    },
    "ImageDescription": {
        "number": 270,
        "type": ctypes.c_char_p,
    },
    "Make": {
        "number": 271,
        "type": ctypes.c_char_p,
    },
    "Model": {
        "number": 272,
        "type": ctypes.c_char_p,
    },
    "StripOffsets": {
        "number": 273,
        "type": (ctypes.c_int32, ctypes.c_int64),
    },
    "Orientation": {
        "number": 274,
        "type": ctypes.c_int16,
    },
    "SamplesPerPixel": {
        "number": 277,
        "type": ctypes.c_int16,
    },
    "RowsPerStrip": {
        "number": 278,
        "type": ctypes.c_int16,
    },
    "StripByteCounts": {
        "number": 279,
        "type": None,
    },
    "MinSampleValue": {
        "number": 280,
        "type": ctypes.c_int16,
    },
    "MaxSampleValue": {
        "number": 281,
        "type": ctypes.c_int16,
    },
    "XResolution": {
        "number": 282,
        "type": ctypes.c_double,
    },
    "YResolution": {
        "number": 283,
        "type": ctypes.c_double,
    },
    "PlanarConfig": {
        "number": 284,
        "type": ctypes.c_int16,
    },
    "PageName": {
        "number": 285,
        "type": ctypes.c_char_p,
    },
    "XPosition": {
        "number": 286,
        "type": ctypes.c_double,
    },
    "YPosition": {
        "number": 287,
        "type": ctypes.c_double,
    },
    "FreeOffsets": {
        "number": 288,
        "type": ctypes.c_int32,
    },
    "FreeByteCounts": {
        "number": 289,
        "type": ctypes.c_int32,
    },
    "GrayResponseUnit": {
        "number": 290,
        "type": ctypes.c_int16,
    },
    "GrayResponseCurve": {
        "number": 291,
        "type": None,
    },
    "T4Options": {
        "number": 292,
        "type": None,
    },
    "T6Options": {
        "number": 293,
        "type": None,
    },
    "ResolutionUnit": {
        "number": 296,
        "type": ctypes.c_int16,
    },
    "PageNumber": {
        "number": 297,
        "type": (ctypes.c_int16, ctypes.c_uint16),
    },
    "TransferFunction": {
        "number": 301,
        "type": None,
    },
    "Software": {
        "number": 305,
        "type": ctypes.c_char_p,
    },
    "Datetime": {
        "number": 306,
        "type": ctypes.c_char_p,
    },
    "Artist": {
        "number": 315,
        "type": ctypes.c_char_p,
    },
    "HostComputer": {
        "number": 316,
        "type": ctypes.c_char_p,
    },
    "Predictor": {
        "number": 317,
        "type": ctypes.c_int16,
    },
    "WhitePoint": {
        "number": 318,
        "type": ctypes.c_double,
    },
    "PrimaryChromaticities": {
        "number": 319,
        "type": None,
    },
    "ColorMap": {
        "number": 320,
        "type": (ctypes.c_int16, ctypes.c_uint16),
    },
    "HalfToneHints": {
        "number": 321,
        "type": ctypes.c_int16,
    },
    "TileWidth": {
        "number": 322,
        "type": ctypes.c_int32,
    },
    "TileLength": {
        "number": 323,
        "type": ctypes.c_int32,
    },
    "TileOffsets": {
        "number": 324,
        "type": None,
    },
    "TileByteCounts": {
        "number": 325,
        "type": None,
    },
    "BadFaxLines": {
        "number": 326,
        "type": None,
    },
    "CleanFaxData": {
        "number": 327,
        "type": None,
    },
    "ConsecutiveBadFaxLines": {
        "number": 328,
        "type": None,
    },
    "SubIFDs": {
        "number": 330,
        "type": None,
    },
    "InkSet": {
        "number": 332,
        "type": ctypes.c_uint16,
    },
    "InkNames": {
        "number": 333,
        "type": ctypes.c_char_p,
    },
    "NumberOfInks": {
        "number": 334,
        "type": ctypes.c_uint16,
    },
    "DotRange": {
        "number": 336,
        "type": None,
    },
    "TargetPrinter": {
        "number": 337,
        "type": ctypes.c_uint16,
    },
    "ExtraSamples": {
        "number": 338,
        "type": ctypes.c_uint16,
    },
    "SampleFormat": {
        "number": 339,
        "type": ctypes.c_uint16,
    },
    "SMinSampleValue": {
        "number": 340,
        "type": ctypes.c_double,
    },
    "SMaxSampleValue": {
        "number": 341,
        "type": ctypes.c_double,
    },
    "TransferRange": {
        "number": 342,
        "type": None,
    },
    "ClipPath": {
        "number": 343,
        "type": None,
    },
    "XClipPathUnits": {
        "number": 344,
        "type": None,
    },
    "YClipPathUnits": {
        "number": 345,
        "type": None,
    },
    "Indexed": {
        "number": 346,
        "type": None,
    },
    "JPEGTables": {
        "number": 347,
        "type": None,
    },
    "OPIProxy": {
        "number": 351,
        "type": None,
    },
    "GlobalParametersIFD": {
        "number": 400,
        "type": None,
    },
    "ProfileType": {
        "number": 401,
        "type": None,
    },
    "FaxProfile": {
        "number": 402,
        "type": ctypes.c_uint8,
    },
    "CodingMethods": {
        "number": 403,
        "type": None,
    },
    "VersionYear": {
        "number": 404,
        "type": None,
    },
    "ModeNumber": {
        "number": 405,
        "type": None,
    },
    "Decode": {
        "number": 433,
        "type": None,
    },
    "DefaultImageColor": {
        "number": 434,
        "type": None,
    },
    "JPEGProc": {
        "number": 512,
        "type": None,
    },
    "JPEGInterchangeFormat": {
        "number": 513,
        "type": None,
    },
    "JPEGInterchangeFormatLength": {
        "number": 514,
        "type": None,
    },
    "JPEGRestartInterval": {
        "number": 515,
        "type": None,
    },
    "JPEGLosslessPredictors": {
        "number": 517,
        "type": None,
    },
    "JPEGPointTransforms": {
        "number": 518,
        "type": None,
    },
    "JPEGQTables": {
        "number": 519,
        "type": None,
    },
    "JPEGDCTables": {
        "number": 520,
        "type": None,
    },
    "JPEGACTables": {
        "number": 521,
        "type": None,
    },
    "YCbCrCoefficients": {
        "number": 529,
        "type": (ctypes.c_float, ctypes.c_float, ctypes.c_float),
    },
    "YCbCrSubsampling": {
        "number": 530,
        "type": (ctypes.c_uint16, ctypes.c_uint16),
    },
    "YCbCrPositioning": {
        "number": 531,
        "type": ctypes.c_uint16,
    },
    "ReferenceBlackWhite": {
        "number": 532,
        "type": (
            ctypes.c_float,
            ctypes.c_float,
            ctypes.c_float,
            ctypes.c_float,
            ctypes.c_float,
            ctypes.c_float,
        ),
    },
    "StripRowCounts": {
        "number": 559,
        "type": None,
    },
    "XMLPacket": {
        "number": 700,
        "type": ctypes.c_uint8,
    },
    "ImageID": {
        "number": 32781,
        "type": None,
    },
    "Rating": {
        "number": 18246,
        "type": ctypes.c_uint16,
    },
    "RatingPercent": {
        "number": 18249,
        "type": ctypes.c_uint16,
    },
    "Datatype": {
        "number": 32996,
        "type": None,
    },
    "WANGAnnotation": {
        "number": 32932,
        "type": None,
    },
    "ImageDepth": {
        "number": 32997,
        "type": None,
    },
    "TileDepth": {
        "number": 32998,
        "type": None,
    },
    "CFARepeatPatternDim": {
        "number": 33421,
        "type": None,
    },
    "CFAPattern": {
        "number": 33422,
        "type": None,
    },
    "BatteryLevel": {
        "number": 33423,
        "type": None,
    },
    "Copyright": {
        "number": 33432,
        "type": ctypes.c_char_p,
    },
    "ExposureTime": {
        "number": 33434,
        "type": ctypes.c_double,
    },
    "FNumber": {
        "number": 33437,
        "type": ctypes.c_double,
    },
    "MDFile": {
        "number": 33445,
        "type": None,
    },
    "MDScalePixel": {
        "number": 33446,
        "type": None,
    },
    "MDColorTable": {
        "number": 33447,
        "type": None,
    },
    "MDLabName": {
        "number": 33448,
        "type": None,
    },
    "MDSampleInfo": {
        "number": 33449,
        "type": None,
    },
    "MdPrepDate": {
        "number": 33450,
        "type": None,
    },
    "MDPrepTime": {
        "number": 33451,
        "type": None,
    },
    "MDFileUnits": {
        "number": 33452,
        "type": None,
    },
    "ModelPixelScale": {
        "number": 33550,
        "type": None,
    },
    "IPTC": {
        "number": 33723,
        "type": None,
    },
    "INGRPacketData": {
        "number": 33918,
        "type": None,
    },
    "INGRFlagRegisters": {
        "number": 33919,
        "type": None,
    },
    "IRASbTransformationMatrix": {
        "number": 33920,
        "type": None,
    },
    "ModelTiePoint": {
        "number": 33922,
        "type": None,
    },
    "ModelTransformation": {
        "number": 34264,
        "type": None,
    },
    "Photoshop": {
        "number": 34377,
        "type": None,
    },
    "ExifTag": {
        "number": 34665,
        "type": ctypes.c_int32,
    },
    "ICCProfile": {
        "number": 34675,
        "type": None,
    },
    "ImageLayer": {
        "number": 34732,
        "type": None,
    },
    "GeoKeyDirectory": {
        "number": 34735,
        "type": None,
    },
    "GeoDoubleParams": {
        "number": 34736,
        "type": None,
    },
    "GeoAsciiParams": {
        "number": 34737,
        "type": None,
    },
    "ExposureProgram": {
        "number": 34850,
        "type": ctypes.c_uint16,
    },
    "SpectralSensitivity": {
        "number": 34852,
        "type": ctypes.c_char_p,
    },
    "GPSIFD": {
        "number": 34853,
        "type": None,
    },
    "ISOSpeedRatings": {
        "number": 34855,
        "type": ctypes.c_uint16,
    },
    "OECF": {
        "number": 34856,
        "type": None,
    },
    "Interlace": {
        "number": 34857,
        "type": None,
    },
    "TimeZoneOffset": {
        "number": 34858,
        "type": None,
    },
    "SelfTimerMode": {
        "number": 34859,
        "type": None,
    },
    "SensitivityType": {
        "number": 34864,
        "type": None,
    },
    "StandardOutputSensitivity": {
        "number": 34865,
        "type": None,
    },
    "RecommendedExposureIndex": {
        "number": 34866,
        "type": None,
    },
    "ISOSpeed": {
        "number": 34867,
        "type": None,
    },
    "ISOSpeedLatitudeYYY": {
        "number": 34868,
        "type": None,
    },
    "ISOSpeedLatitudeZZZ": {
        "number": 34869,
        "type": None,
    },
    "HYLAFAXRecvParams": {
        "number": 34908,
        "type": None,
    },
    "HYLAFAXSubAddress": {
        "number": 34909,
        "type": None,
    },
    "HYLAFAXRecvTime": {
        "number": 34910,
        "type": None,
    },
    "ExifVersion": {
        "number": 36864,
        "type": ctypes.c_uint8,
    },
    "DateTimeOriginal": {
        "number": 36867,
        "type": None,
    },
    "DateTimeDigitized": {
        "number": 36868,
        "type": None,
    },
    "OffsetTime": {
        "number": 36880,
        "type": None,
    },
    "OffsetTimeOriginal": {
        "number": 36881,
        "type": None,
    },
    "OffsetTimeDigitized": {
        "number": 36882,
        "type": None,
    },
    "ComponentsConfiguration": {"number": 37121, "type": None},
    "CompressedBitsPerPixel": {
        "number": 37122,
        "type": ctypes.c_uint8,
    },
    "ShutterSpeedValue": {
        "number": 37377,
        "type": ctypes.c_double,
    },
    "ApertureValue": {
        "number": 37378,
        "type": ctypes.c_double,
    },
    "BrightnessValue": {
        "number": 37379,
        "type": ctypes.c_double,
    },
    "ExposureBiasValue": {
        "number": 37380,
        "type": ctypes.c_double,
    },
    "MaxApertureValue": {
        "number": 37381,
        "type": ctypes.c_double,
    },
    "SubjectDistance": {
        "number": 37382,
        "type": ctypes.c_double,
    },
    "MeteringMode": {
        "number": 37383,
        "type": ctypes.c_uint16,
    },
    "LightSource": {
        "number": 37384,
        "type": ctypes.c_uint16,
    },
    "Flash": {
        "number": 37385,
        "type": ctypes.c_uint16,
    },
    "FocalLength": {
        "number": 37386,
        "type": ctypes.c_double,
    },
    "Noise": {
        "number": 37389,
        "type": None,
    },
    "ImageNumber": {
        "number": 37393,
        "type": None,
    },
    "SecurityClassification": {
        "number": 37394,
        "type": None,
    },
    "ImageHistory": {
        "number": 37395,
        "type": None,
    },
    "TIFFEPStandardID": {
        "number": 37398,
        "type": None,
    },
    "MakerNote": {
        "number": 37500,
        "type": ctypes.c_char_p,
    },
    "UserComment": {
        "number": 37510,
        "type": ctypes.c_char_p,
    },
    "SubSecTime": {
        "number": 37520,
        "type": ctypes.c_char_p,
    },
    "SubSecTimeOriginal": {
        "number": 37521,
        "type": ctypes.c_char_p,
    },
    "SubSecTimeDigitized": {
        "number": 37522,
        "type": ctypes.c_char_p,
    },
    "ImageSourceData": {
        "number": 37724,
        "type": None,
    },
    "Temperature": {
        "number": 37888,
        "type": None,
    },
    "Humidity": {
        "number": 37889,
        "type": None,
    },
    "Pressure": {
        "number": 37890,
        "type": None,
    },
    "WaterDepth": {
        "number": 37891,
        "type": None,
    },
    "Acceleration": {
        "number": 37892,
        "type": None,
    },
    "CameraElevationAngle": {
        "number": 37893,
        "type": None,
    },
    "XPTitle": {
        "number": 40091,
        "type": None,
    },
    "XPComment": {
        "number": 40092,
        "type": None,
    },
    "XPAuthor": {
        "number": 40093,
        "type": None,
    },
    "XPKeywords": {
        "number": 40094,
        "type": None,
    },
    "XPSubject": {
        "number": 40095,
        "type": None,
    },
    "FlashPixVersion": {
        "number": 40960,
        "type": None,
    },
    "ColorSpace": {
        "number": 40961,
        "type": ctypes.c_uint16,
    },
    "PixelXDimension": {
        "number": 40962,
        "type": ctypes.c_uint64,
    },
    "PixelYDimension": {
        "number": 40963,
        "type": ctypes.c_uint64,
    },
    "InteroperabilityIFD": {
        "number": 40965,
        "type": None,
    },
    "FocalPlaneXResolution": {
        "number": 41486,
        "type": ctypes.c_double,
    },
    "FocalPlaneYResolution": {
        "number": 41487,
        "type": ctypes.c_double,
    },
    "FocalPlaneResolutionUnit": {
        "number": 41488,
        "type": ctypes.c_uint16,
    },
    "ExposureIndex": {
        "number": 41493,
        "type": ctypes.c_double,
    },
    "SensingMethod": {
        "number": 41495,
        "type": ctypes.c_uint16,
    },
    "SceneType": {
        "number": 41729,
        "type": ctypes.c_uint8,
    },
    "FlashEnergy": {"number": 41483, "type": None},
    "SpatialFrequencyResponse": {"number": 41484, "type": None},
    "SubjectLocation": {"number": 41492, "type": None},
    "FileSource": {"number": 41728, "type": None},
    "ExifCFAPattern": {"number": 41730, "type": None},
    "CustomRendered": {"number": 41985, "type": None},
    "ExposureMode": {
        "number": 41986,
        "type": ctypes.c_uint16,
    },
    "WhiteBalance": {
        "number": 41987,
        "type": ctypes.c_uint16,
    },
    "DigitalZoomRatio": {
        "number": 41988,
        "type": ctypes.c_double,
    },
    "FocalLengthIn35mmFilm": {
        "number": 41989,
        "type": ctypes.c_uint16,
    },
    "SceneCaptureType": {
        "number": 41990,
        "type": ctypes.c_uint16,
    },
    "GainControl": {
        "number": 41991,
        "type": ctypes.c_uint16,
    },
    "Contrast": {
        "number": 41992,
        "type": ctypes.c_uint16,
    },
    "Saturation": {
        "number": 41993,
        "type": ctypes.c_uint16,
    },
    "Sharpness": {
        "number": 41994,
        "type": ctypes.c_uint16,
    },
    "DeviceSettingDescription": {
        "number": 41995,
        "type": ctypes.c_char_p,
    },
    "SubjectDistanceRange": {
        "number": 41996,
        "type": ctypes.c_uint16,
    },
    "ImageUniqueID": {"number": 42016, "type": None},
    "CameraOwnerName": {"number": 42032, "type": None},
    "BodySerialNumber": {"number": 42033, "type": None},
    "LensSpecification": {"number": 42034, "type": None},
    "LensMake": {"number": 42035, "type": None},
    "LensModel": {"number": 42036, "type": None},
    "LensSerialNumber": {"number": 42037, "type": None},
    "CompositeImage": {"number": 42080, "type": None},
    "SourceImageNumberOfCompositeImage": {"number": 42081, "type": None},
    "SourceExposureTimeOfCompositeImage": {"number": 42082, "type": None},
    "GDAL_Metadata": {
        "number": 42112,
        "type": None,
    },
    "GDAL_NoData": {
        "number": 42113,
        "type": None,
    },
    "Gamma": {
        "number": 42240,
        "type": None,
    },
    "OCEScanJobDescription": {
        "number": 50215,
        "type": None,
    },
    "OCEApplicationSelector": {
        "number": 50216,
        "type": None,
    },
    "OCEIdentificationNumber": {
        "number": 50217,
        "type": None,
    },
    "OCEImageLogicCharacteristics": {
        "number": 50218,
        "type": None,
    },
    "PrintImageMatching": {
        "number": 50341,
        "type": None,
    },
    "DNGVersion": {
        "number": 50706,
        "type": None,
    },
    "DNGBackwardVersion": {
        "number": 50707,
        "type": None,
    },
    "UniqueCameraModel": {
        "number": 50708,
        "type": None,
    },
    "LocalizedCameraModel": {
        "number": 50709,
        "type": None,
    },
    "CFAPlaneColor": {
        "number": 50710,
        "type": None,
    },
    "CFALayout": {
        "number": 50711,
        "type": None,
    },
    "LinearizationTable": {
        "number": 50712,
        "type": None,
    },
    "BlackLevelRepeatDim": {
        "number": 50713,
        "type": None,
    },
    "BlackLevel": {
        "number": 50714,
        "type": None,
    },
    "BlackLevelDeltaH": {
        "number": 50715,
        "type": None,
    },
    "BlackLevelDeltaV": {
        "number": 50716,
        "type": None,
    },
    "WhiteLevel": {
        "number": 50717,
        "type": None,
    },
    "DefaultScale": {
        "number": 50718,
        "type": None,
    },
    "DefaultCropOrigin": {
        "number": 50719,
        "type": None,
    },
    "DefaultCropSize": {
        "number": 50720,
        "type": None,
    },
    "ColorMatrix1": {
        "number": 50721,
        "type": None,
    },
    "ColorMatrix2": {
        "number": 50722,
        "type": None,
    },
    "CameraCalibration1": {
        "number": 50723,
        "type": None,
    },
    "CameraCalibration2": {
        "number": 50724,
        "type": None,
    },
    "ReductionMatrix1": {
        "number": 50725,
        "type": None,
    },
    "ReductionMatrix2": {
        "number": 50726,
        "type": None,
    },
    "AnalogBalance": {
        "number": 50727,
        "type": None,
    },
    "AsShotNeutral": {
        "number": 50728,
        "type": None,
    },
    "AsShotWhiteXY": {
        "number": 50729,
        "type": None,
    },
    "BaselineExposure": {
        "number": 50730,
        "type": None,
    },
    "BaselineNoise": {
        "number": 50731,
        "type": None,
    },
    "BaselineSharpness": {
        "number": 50732,
        "type": None,
    },
    "BayerGreenSplit": {
        "number": 50733,
        "type": None,
    },
    "LinearResponseLimit": {
        "number": 50734,
        "type": None,
    },
    "CameraSerialNumber": {
        "number": 50735,
        "type": None,
    },
    "LensInfo": {
        "number": 50736,
        "type": None,
    },
    "ChromaBlurRadius": {
        "number": 50737,
        "type": None,
    },
    "AntiAliasStrength": {
        "number": 50738,
        "type": None,
    },
    "ShadowScale": {
        "number": 50739,
        "type": None,
    },
    "DNGPrivateData": {
        "number": 50740,
        "type": None,
    },
    "MakerNoteSafety": {
        "number": 50741,
        "type": None,
    },
    "CalibrationIllumintant1": {
        "number": 50778,
        "type": None,
    },
    "CalibrationIllumintant2": {
        "number": 50779,
        "type": None,
    },
    "BestQualityScale": {
        "number": 50780,
        "type": None,
    },
    "RawDataUniqueID": {
        "number": 50781,
        "type": None,
    },
    "AliasLayerMetadata": {
        "number": 50784,
        "type": None,
    },
    "OriginalRawFileName": {
        "number": 50827,
        "type": None,
    },
    "OriginalRawFileData": {
        "number": 50828,
        "type": None,
    },
    "ActiveArea": {
        "number": 50829,
        "type": None,
    },
    "MaskedAreas": {
        "number": 50830,
        "type": None,
    },
    "AsShotICCProfile": {
        "number": 50831,
        "type": None,
    },
    "AsShotPreProfileMatrix": {
        "number": 50832,
        "type": None,
    },
    "CurrentICCProfile": {
        "number": 50833,
        "type": None,
    },
    "CurrentPreProfileMatrix": {
        "number": 50834,
        "type": None,
    },
    "ColorimetricReference": {
        "number": 50839,
        "type": None,
    },
    "TIFF_RSID": {
        "number": 50908,
        "type": None,
    },
    "GEO_Metadata": {
        "number": 50909,
        "type": None,
    },
    "CameraCalibrationSignature": {
        "number": 50931,
        "type": None,
    },
    "ProfileCalibrationSignature": {
        "number": 50932,
        "type": None,
    },
    "AsShotProfileName": {
        "number": 50934,
        "type": None,
    },
    "NoiseReductionApplied": {
        "number": 50935,
        "type": None,
    },
    "ProfileName": {
        "number": 50936,
        "type": None,
    },
    "ProfileHueSatMapDims": {
        "number": 50937,
        "type": None,
    },
    "ProfileHueSatMapData1": {
        "number": 50938,
        "type": None,
    },
    "ProfileHueSatMapData2": {
        "number": 50939,
        "type": None,
    },
    "ProfileToneCurve": {
        "number": 50940,
        "type": None,
    },
    "ProfileEmbedPolicy": {
        "number": 50941,
        "type": None,
    },
    "ProfileCopyright": {
        "number": 50942,
        "type": None,
    },
    "ForwardMatrix1": {
        "number": 50964,
        "type": None,
    },
    "ForwardMatrix2": {
        "number": 50965,
        "type": None,
    },
    "PreviewApplicationName": {
        "number": 50966,
        "type": None,
    },
    "PreviewApplicationVersion": {
        "number": 50967,
        "type": None,
    },
    "PreviewSettingsName": {
        "number": 50968,
        "type": None,
    },
    "PreviewSettingsDigest": {
        "number": 50969,
        "type": None,
    },
    "PreviewColorSpace": {
        "number": 50970,
        "type": None,
    },
    "PreviewDateTime": {
        "number": 50971,
        "type": None,
    },
    "RawImageDigest": {
        "number": 50972,
        "type": None,
    },
    "OriginalRawFileDigest": {
        "number": 50973,
        "type": None,
    },
    "SubTileBlockSize": {
        "number": 50974,
        "type": None,
    },
    "RowInterleaveFactor": {
        "number": 50975,
        "type": None,
    },
    "ProfileLookTableDims": {
        "number": 50981,
        "type": None,
    },
    "ProfileLookTableData": {
        "number": 50982,
        "type": None,
    },
    "OpcodeList1": {
        "number": 51008,
        "type": None,
    },
    "OpcodeList2": {
        "number": 51009,
        "type": None,
    },
    "OpcodeList3": {
        "number": 51022,
        "type": None,
    },
    "NoiseProfile": {
        "number": 51041,
        "type": None,
    },
    "JPEGQuality": {
        "number": 65537,
        "type": ctypes.c_int32,
    },
    "JPEGColorMode": {
        "number": 65538,
        "type": ctypes.c_int32,
    },
}

# We need the reverse mapping as well.
TAGNUM2NAME = {value["number"]: key for key, value in TAGS.items()}


def tiff_header(read_buffer):
    """Interpret the uuid raw data as a tiff header."""
    # First 8 should be (73, 73, 42, 8) or (77, 77, 42, 8)
    data = struct.unpack("BB", read_buffer[0:2])
    if data[0] == 73 and data[1] == 73:
        # little endian
        endian = "<"
    elif data[0] == 77 and data[1] == 77:
        # big endian
        endian = ">"
    else:
        msg = (
            f"The byte order indication in the TIFF header "
            f"({read_buffer[0:2]}) is invalid.  It should be either "
            f"{bytes([73, 73])} or {bytes([77, 77])}."
        )
        raise RuntimeError(msg)

    _, offset = struct.unpack(endian + "HI", read_buffer[2:8])

    # This is the 'Exif Image' portion.
    return _Ifd(endian, read_buffer, offset).processed_ifd


class BadTiffTagDatatype(RuntimeError):
    """This exception exists soley to better communicate up the stack that the
    problem exists.
    """

    pass


class _Ifd(object):
    """Corresponds to TIFF IFD data structure on file.

    Attributes
    ----------
    read_buffer : bytes
        Raw byte stream consisting of the UUID data.
    endian : str
        Either '<' for big-endian, or '>' for little-endian.
    num_tags : int
        Number of tags in the IFD.
    raw_ifd : dictionary
        Maps tag number to "mildly-interpreted" tag value.
    processed_ifd : dictionary
        Maps tag name to "mildly-interpreted" tag value.
    """

    def __init__(self, endian, read_buffer, offset):
        self.endian = endian
        self.read_buffer = read_buffer
        self.processed_ifd = OrderedDict()

        (self.num_tags,) = struct.unpack(
            endian + "H",
            read_buffer[offset:offset + 2]
        )

        fmt = self.endian + "HHII" * self.num_tags
        ifd_buffer = read_buffer[offset + 2:offset + 2 + self.num_tags * 12]
        data = struct.unpack(fmt, ifd_buffer)
        self.raw_ifd = OrderedDict()
        for j, tag in enumerate(data[0::4]):
            # The offset to the tag offset/payload is the offset to the IFD
            # plus 2 bytes for the number of tags plus 12 bytes for each
            # tag entry plus 8 bytes to the offset/payload itself.
            toffp = read_buffer[offset + 10 + j * 12:offset + 10 + j * 12 + 4]
            self.raw_ifd[tag] = self.parse_tag(
                tag, data[j * 4 + 1], data[j * 4 + 2], toffp
            )

        self.post_process()

    def parse_tag(self, tag, dtype, count, offset_buf):
        """Interpret an Exif image tag data payload."""

        try:
            fmt = DATATYPE2FMT[dtype]["format"] * count
            payload_size = DATATYPE2FMT[dtype]["nbytes"] * count
        except KeyError:
            msg = f"Invalid TIFF tag datatype ({dtype})."
            raise BadTiffTagDatatype(msg)

        if payload_size <= 4:
            # Interpret the payload from the 4 bytes in the tag entry.
            target_buffer = offset_buf[:payload_size]
        else:
            # Interpret the payload at the offset specified by the 4 bytes in
            # the tag entry.
            (offset,) = struct.unpack(self.endian + "I", offset_buf)
            target_buffer = self.read_buffer[offset:offset + payload_size]

        if dtype == 2:
            # ASCII
            payload = target_buffer.decode("utf-8").rstrip("\x00")

        else:
            payload = struct.unpack(self.endian + fmt, target_buffer)
            if dtype == 5 or dtype == 10:
                # Rational or Signed Rational.  Construct the list of values.
                rational_payload = []
                for j in range(count):
                    try:
                        value = float(payload[j * 2]) / float(payload[j * 2 + 1])  # noqa : E501
                    except ZeroDivisionError:
                        value = np.nan
                    rational_payload.append(value)
                payload = np.array(rational_payload)
            if count == 1:
                # If just a single value, then return a scalar instead of a
                # tuple.
                payload = payload[0]
            else:
                payload = np.array(payload, dtype=DATATYPE2FMT[dtype]["nptype"])

        return payload

    def post_process(self):
        """Map the tag name instead of tag number to the tag value."""
        for tag, value in self.raw_ifd.items():
            try:
                tag_name = TAGNUM2NAME[tag]
            except KeyError:
                # Ok, we don't recognize this tag.  Just use the numeric id.
                msg = f"Unrecognized UUID box TIFF tag ({tag})."
                warnings.warn(msg, UserWarning)
                tag_name = tag

            if tag_name in ("ExifTag", 'GPSIFD'):
                # There's an IFD at the offset specified here.
                ifd = _Ifd(self.endian, self.read_buffer, value)
                self.processed_ifd[tag_name] = ifd.processed_ifd
            else:
                # just a regular tag, treat it as a simple value
                self.processed_ifd[tag_name] = value


# maps the TIFF enumerated datatype to the corresponding structs datatype code,
# the data width, and the corresponding numpy datatype
DATATYPE2FMT = {
    1: {"format": "B", "nbytes": 1, "nptype": np.ubyte},
    2: {"format": "B", "nbytes": 1, "nptype": str},
    3: {"format": "H", "nbytes": 2, "nptype": np.uint16},
    4: {"format": "I", "nbytes": 4, "nptype": np.uint32},
    5: {"format": "II", "nbytes": 8, "nptype": np.double},
    6: {"format": "b", "nbytes": 1, "nptype": np.int8},
    7: {"format": "B", "nbytes": 1, "nptype": np.uint8},
    8: {"format": "h", "nbytes": 2, "nptype": np.int16},
    9: {"format": "i", "nbytes": 4, "nptype": np.int32},
    10: {"format": "ii", "nbytes": 8, "nptype": np.double},
    11: {"format": "f", "nbytes": 4, "nptype": np.double},
    12: {"format": "d", "nbytes": 8, "nptype": np.double},
    13: {"format": "I", "nbytes": 4, "nptype": np.uint32},
    16: {"format": "Q", "nbytes": 8, "nptype": np.uint64},
    17: {"format": "q", "nbytes": 8, "nptype": np.int64},
    18: {"format": "Q", "nbytes": 8, "nptype": np.uint64},
}



# standard library imports
from __future__ import annotations
import logging
import pathlib
import shutil
import sys
from typing import List, Tuple
import warnings

# 3rd party library imports
import numpy as np

# local imports
from glymur import Jp2k, set_option
from glymur.core import SRGB
from ._core_converter import _2JP2Converter
from .lib import tiff as libtiff
from . import jp2box

# we need a lower case mapping from the tag name to the tag number
TAGNAME2NUM = {k.lower(): v["number"] for k, v in libtiff.TAGS.items()}


class Tiff2Jp2k(_2JP2Converter):
    """
    Transform a TIFF image into a JP2 image.

    Attributes
    ----------
    create_exif_uuid : bool
        Create a UUIDBox for the TIFF IFD metadata.
    dtype : np.dtype
        Datatype of the image.
    found_geotiff_tags : bool
        If true, then this TIFF must be a GEOTIFF
    imageheight, imagewidth : int
        Dimensions of the image.
    jp2 : JP2K object
        Write to this JPEG2000 file
    jp2_path : path
        Path to JPEG 2000 file to be written.
    jp2_kwargs : dict
        Keyword arguments to pass along to the Jp2k constructor.
    photo : int
        The photometric interpretation of the image.
    rps : int
        The number of rows per strip in the TIFF.
    spp : int
        Samples Per Pixel TIFF tag value
    tiff_path : path
        Path to TIFF file.
    tilesize : tuple
        The dimensions of a tile in the JP2K file.
    tw, th : int
        The tile dimensions for the TIFF image.
    version : int
        Identifies the TIFF as 32-bit TIFF or 64-bit TIFF.
    xmp_data : bytes
        Encoded bytes from XML_PACKET tag (700), or None if not present.
    """

    def __init__(
        self,
        tiff_path: pathlib.Path,
        jp2_path: pathlib.Path,
        create_exif_uuid: bool = True,
        create_xmp_uuid: bool = True,
        exclude_tags: List[int | str] | None = None,
        num_threads: int = 1,
        tilesize: Tuple[int, int] | None = None,
        include_icc_profile: bool = False,
        verbosity: int = logging.CRITICAL,
        **kwargs,
    ):
        """
        Construct the object.

        Parameters
        ----------
        create_exif_uuid : bool
            If true, create an EXIF UUID out of the TIFF metadata (tags)
        create_xmp_uuid : bool
            If true and if there is an XMLPacket (700) tag in the TIFF main
            IFD, it will be removed from the IFD and placed in a UUID box.
        include_icc_profile : bool
            If true, consume any ICC profile tag (34765) into the colour
            specification box.
        exclude_tags : list or None
            If not None and if create_exif_uuid is True, exclude any listed
            tags from the EXIF UUID.
        jp2_filename : path or str
            Path to JPEG 2000 file to be written.
        tiff_filename : path or str
            Path to TIFF file.
        tilesize : tuple
            The dimensions of a tile in the JP2K file.
        verbosity : int
            Set the level of logging, i.e. WARNING, INFO, etc.
        """
        super().__init__(
            create_exif_uuid, create_xmp_uuid, include_icc_profile, tilesize,
            verbosity
        )

        self.tiff_path = pathlib.Path(tiff_path)
        if not self.tiff_path.exists():
            raise FileNotFoundError(f"{tiff_path} does not exist")

        self.jp2_path = pathlib.Path(jp2_path)
        if self.jp2_path.exists():
            msg = (
                f'{str(self.jp2_path)} already exists, ',
                'please delete if you wish to overwrite.'
            )
            raise FileExistsError(msg)

        self.create_exif_uuid = create_exif_uuid
        self.create_xmp_uuid = create_xmp_uuid

        if exclude_tags is None:
            exclude_tags = []
        self.exclude_tags = self._process_exclude_tags(exclude_tags)

        self.jp2 = None
        self.jp2_kwargs = kwargs

        # Assume that there is no ColorMap tag until we know otherwise.
        self._colormap = None

        # Assume no XML_PACKET tag until we know otherwise.
        self.xmp_data = None

        self.setup_logging(verbosity)

        if num_threads > 1:
            set_option("lib.num_threads", num_threads)

    def _process_exclude_tags(self, exclude_tags):
        """The list of tags to exclude may be mixed type (str or integer).
        There is also the possibility that they may be capitalized differently
        than our internal list, so the goal here is to convert them all to
        integer values.

        Parameters
        ----------
        exclude_tags : list
            List of tags that are meant to be excluded from the EXIF UUID.

        Returns
        -------
        list of numeric tag values
        """
        lst = []

        # first, make the tags all str datatype
        # compare the tags as lower case for consistency's sake
        exclude_tags = [
            tag.lower() if isinstance(tag, str) else str(tag)
            for tag in exclude_tags
        ]

        # If any tags were specified as strings, we need to convert them
        # into tag numbers.
        for tag in exclude_tags:

            # convert from str to numeric
            #
            # is it a string like '325'?  then convert to integer 325
            try:
                tag_num = int(tag)
            except ValueError:
                # tag wasn't '325', but rather 'tilebytecounts',
                # hopefully?  Try to map from the name back to the tag
                # number.
                try:
                    tag_num = TAGNAME2NUM[tag]
                except KeyError:
                    msg = f"{tag} is not a recognized TIFF tag"
                    warnings.warn(msg)
                else:
                    lst.append(tag_num)
            else:
                # tag really was something like '325', so we keep the
                # numeric value
                lst.append(tag_num)

        return lst

    def __enter__(self):
        """The Tiff2Jp2k must be used with a context manager."""
        self.tiff_fp = libtiff.open(str(self.tiff_path))
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        libtiff.close(self.tiff_fp)

    def run(self):

        self.get_main_ifd()
        self.copy_image()
        self.append_extra_jp2_boxes()
        self.rewrap_jp2()

    def rewrap_jp2(self):
        """These re-wrap operations should be mutually exclusive.  An ICC
        profile should not exist in a TIFF with a colormap.
        """
        self.rewrap_for_colormap()
        self.rewrap_for_icc_profile()

    def rewrap_for_colormap(self):
        """If the photometric interpretation was PALETTE, then we need to
        insert a pclr box and a cmap (component mapping box).
        """

        photo = self.get_tag_value(262)
        if photo != libtiff.Photometric.PALETTE:
            return

        jp2h = [box for box in self.jp2.box if box.box_id == "jp2h"][0]

        bps = (8, 8, 8)
        pclr = jp2box.PaletteBox(
            palette=self._colormap,
            bits_per_component=bps,
            signed=(False, False, False)
        )
        jp2h.box.append(pclr)

        # append the component mapping box
        cmap = jp2box.ComponentMappingBox(
            component_index=(0, 0, 0),
            mapping_type=(1, 1, 1),
            palette_index=(0, 1, 2)
        )
        jp2h.box.append(cmap)

        # fix the colr box.  the colorspace needs to be changed from greyscale
        # to rgb
        colr = [box for box in jp2h.box if box.box_id == "colr"][0]
        colr.colorspace = SRGB

        temp_filename = str(self.jp2_path) + ".tmp"
        self.jp2.wrap(temp_filename, boxes=self.jp2.box)
        shutil.move(temp_filename, self.jp2_path)
        self.jp2.parse()

    def append_extra_jp2_boxes(self):
        """Copy over the TIFF IFD.  Place it in a UUID box.  Append to the JPEG
        2000 file.
        """
        self.append_exif_uuid_box()
        self.append_xmp_uuid_box()

    def get_main_ifd(self):
        """Read all the tags in the main IFD.  We do it this way because of the
        difficulty in using TIFFGetFieldDefaulted when the datatype of a tag
        can differ.
        """

        with self.tiff_path.open(mode="rb") as tfp:

            self.read_tiff_header(tfp)

            self.tags = self.read_ifd(tfp)

            if 320 in self.tags:

                # the TIFF must have PALETTE photometric interpretation
                data = np.array(self.tags[320]["payload"])
                self._colormap = data.reshape(len(data) // 3, 3)
                self._colormap = self._colormap / 65535
                self._colormap = (self._colormap * 255).astype(np.uint8)

            if 700 in self.tags:

                # XMLPacket
                self.xmp_data = bytes(self.tags[700]["payload"])

            else:
                self.xmp_data = None

            if 34675 in self.tags:
                # ICC profile
                self.icc_profile = bytes(self.tags[34675]["payload"])

            else:
                self.icc_profile = None

    def get_tag_value(self, tagnum):
        """Return the value associated with the tag.  Some tags are not
        actually written into the IFD, but are instead "defaulted".

        Returns
        -------
        tag value
        """

        if tagnum not in self.tags and tagnum == 284:
            # PlanarConfig is not always written into the IFD, defaults to 1
            return 1

        if tagnum not in self.tags and tagnum == 339:
            # SampleFormat is not always written into the IFD, defaults to 1
            return 1

        # The tag value is always stored as a tuple with at least one member.
        return self.tags[tagnum]["payload"][0]

    def copy_image(self):
        """Transfer the image data from the TIFF to the JPEG 2000 file."""

        if libtiff.isTiled(self.tiff_fp):
            isTiled = True
        else:
            isTiled = False

        self.photo = self.get_tag_value(262)
        self.imagewidth = self.get_tag_value(256)
        self.imageheight = self.get_tag_value(257)
        self.spp = self.get_tag_value(277)
        sf = self.get_tag_value(339)
        bps = self.get_tag_value(258)
        planar = self.get_tag_value(284)

        if sf not in [libtiff.SampleFormat.UINT, libtiff.SampleFormat.VOID]:
            sampleformat_str = self.tagvalue2str(libtiff.SampleFormat, sf)

            msg = (
                f"The TIFF SampleFormat is {sampleformat_str}.  Only UINT "
                "and VOID are supported."
            )
            raise RuntimeError(msg)

        if bps == 8 and sf == libtiff.SampleFormat.UINT:
            self.dtype = np.uint8
        elif bps == 16 and sf == libtiff.SampleFormat.UINT:
            self.dtype = np.uint16
        else:
            msg = (
                f"Only unsigned sample formats and bits per sample values of "
                f"8 or 16 are supported.  The values in the TIFF are {bps} "
                f"{sf}."
            )
            raise RuntimeError(msg)

        if (
            planar == libtiff.PlanarConfig.SEPARATE
            and self.tilesize is not None
        ):
            msg = (
                "A separated planar configuration is not supported when a "
                "tile size is specified."
            )
            raise RuntimeError(msg)

        if libtiff.isTiled(self.tiff_fp):
            self.tw = self.get_tag_value(322)
            self.th = self.get_tag_value(323)
        else:
            self.tw = self.imagewidth
            try:
                self.rps = self.get_tag_value(278)
            except KeyError:
                # stripped but no RowsPerStrip tag?  default to the image
                # height
                self.rps = self.imageheight

        if self.spp == 1:
            shape = (self.imageheight, self.imagewidth)
        else:
            shape = (self.imageheight, self.imagewidth, self.spp)

        self.jp2 = Jp2k(
            self.jp2_path,
            shape=shape,
            tilesize=self.tilesize,
            **self.jp2_kwargs
        )

        if not libtiff.RGBAImageOK(self.tiff_fp):
            photometric_string = self.tagvalue2str(
                libtiff.Photometric, self.photo
            )
            msg = (
                f"The TIFF Photometric tag is {photometric_string}.  It is "
                "not supported by this program."
            )
            raise RuntimeError(msg)
        elif self.tilesize is None and self.photo == libtiff.Photometric.YCBCR:
            # this handles both YCbCr cases of a striped TIFF and a tiled TIFF
            self._write_rgba_single_tile()
            self.jp2.finalize(force_parse=True)
        elif self.tilesize is None and isTiled:
            self._write_tiled_tiff_to_single_tile_jp2k()
        elif self.tilesize is None and not isTiled:
            self._write_stripped_tiff_to_single_tile_jp2k()
        elif isTiled and self.tilesize is not None:
            self._write_tiled_tiff_to_tiled_jp2k()
        elif not isTiled and self.tilesize is not None:
            self._write_striped_tiff_to_tiled_jp2k()

    def tagvalue2str(self, cls, tag_value):
        """Take a class that encompasses all of a tag's allowed values and find
        the name of that value.
        """

        tag_value_string = [
            key for key in dir(cls) if getattr(cls, key) == tag_value
        ][0]

        return tag_value_string

    def _write_rgba_single_tile(self):
        """If no jp2k tiling was specified and if the image is ok to read
        via the RGBA interface, then just do that.  The image will be
        written with the tilesize equal to the image size, so it will
        be written using a single write operation.
        """
        msg = (
            "Reading using the RGBA interface, writing as a single tile "
            "image."
        )
        self.logger.info(msg)

        image = libtiff.readRGBAImageOriented(
            self.tiff_fp, self.imagewidth, self.imageheight
        )

        # must reorder image planes on big-endian
        if sys.byteorder == "big":
            image = np.flip(image, axis=2)

        # potentially get rid of the alpha plane
        if self.spp < 4:
            image = image[:, :, :3]

        self.jp2[:] = image

    def _write_stripped_tiff_to_single_tile_jp2k(self):
        """The input TIFF image is stripped and we are to create the output
        JPEG2000 image as a single tile.
        """
        num_tiff_strip_rows = int(np.ceil(self.imageheight / self.rps))

        # This might be a bit bigger than the actual image because of a
        # possibly partial last strip.
        stripped_shape = (
            num_tiff_strip_rows * self.rps,
            self.imagewidth,
            self.spp
        )
        image = np.zeros(stripped_shape, dtype=self.dtype)

        tiff_strip = np.zeros(
            (self.rps, self.imagewidth, self.spp), dtype=self.dtype
        )

        # manually collect all the tiff strips, stuff into the image
        for stripnum in range(num_tiff_strip_rows):
            rows = slice(stripnum * self.rps, (stripnum + 1) * self.rps)
            libtiff.readEncodedStrip(self.tiff_fp, stripnum, tiff_strip)
            image[rows, :, :] = tiff_strip

        if self.imageheight != stripped_shape[0]:
            # cut the image down due to a partial last strip
            image = image[: self.imageheight, :, :]

        self.jp2[:] = image

    def _write_tiled_tiff_to_single_tile_jp2k(self):
        """The input TIFF image is tiled and we are to create the output
        JPEG2000 image as a single tile.
        """
        num_tiff_tile_cols = int(np.ceil(self.imagewidth / self.tw))
        num_tiff_tile_rows = int(np.ceil(self.imageheight / self.th))

        # tiled shape might differ from the final image shape if we have
        # partial tiles on the bottom and on the right
        final_shape = self.imageheight, self.imagewidth, self.spp
        tiled_shape = (
            num_tiff_tile_rows * self.th,
            num_tiff_tile_cols * self.tw,
            self.spp,
        )

        image = np.zeros(tiled_shape, dtype=self.dtype)
        tiff_tile = np.zeros((self.th, self.tw, self.spp), dtype=self.dtype)

        # manually collect all the tiff tiles, stuff into the image
        for tr in range(num_tiff_tile_rows):

            rows = slice(tr * self.th, (tr + 1) * self.th)

            for tc in range(num_tiff_tile_cols):
                ttile_num = num_tiff_tile_cols * tr + tc
                libtiff.readEncodedTile(self.tiff_fp, ttile_num, tiff_tile)

                cols = slice(tc * self.tw, (tc + 1) * self.tw)

                image[rows, cols, :] = tiff_tile

        if final_shape != tiled_shape:
            image = image[: final_shape[0], : final_shape[1], :]

        self.jp2[:] = image

    def _write_tiled_tiff_to_tiled_jp2k(self):
        """The input TIFF image is tiled and we are to create the output
        JPEG2000 image with specific tile dimensions.
        """
        for jp2k_tilenum, tilewriter in enumerate(self.jp2.get_tilewriters()):
            tiff_tiles = self._get_covering_tiles(jp2k_tilenum)
            jp2k_tile = self._cover_tile(jp2k_tilenum, tiff_tiles)
            self.logger.info(f"Writing tile {jp2k_tilenum}")
            tilewriter[:] = jp2k_tile

    def _cover_tile(self, jp2k_tile_num, tiff_tile_nums):
        """
        Fill in the jp2k tile with image data from the TIFF tiles.

        Parameters
        ----------
        jp2k_tilenum : int
            number of the JPEG2000 tile
        tiff_tile_nums : set
            all TIFF tile identifiers that cover the jpeg2000 tile
        """
        jth, jtw = self.tilesize

        # Does the JP2K have partial tiles on the far right and bottom of the
        # image.
        partial_jp2_tile_rows = (self.imageheight / jth) != (
            self.imageheight // jth
        )  # noqa : E501
        partial_jp2_tile_cols = (self.imagewidth / jtw) != (
            self.imagewidth // jtw
        )  # noqa : E501

        num_jp2k_tile_rows = int(np.ceil(self.imageheight / jth))
        num_jp2k_tile_cols = int(np.ceil(self.imagewidth / jtw))

        jp2k_tile_row = int(np.ceil(jp2k_tile_num // num_jp2k_tile_cols))
        jp2k_tile_col = int(np.ceil(jp2k_tile_num % num_jp2k_tile_cols))

        num_tiff_tile_cols = int(np.ceil(self.imagewidth / self.tw))

        jp2k_tile = np.zeros((jth, jtw, self.spp), dtype=self.dtype)

        # coordinates of jp2k upper left corner
        jp2k_ulx = jtw * jp2k_tile_col
        jp2k_uly = jth * jp2k_tile_row

        for ttile_num in tiff_tile_nums:

            # compute the coordinates of the upper left pixel of the TIFF tile
            tiff_tile_row = int(np.ceil(ttile_num // num_tiff_tile_cols))
            tiff_tile_col = int(np.ceil(ttile_num % num_tiff_tile_cols))
            tile_uly = tiff_tile_row * self.th
            tile_ulx = tiff_tile_col * self.tw

            if self.photo == libtiff.Photometric.YCBCR:

                rgba_tile = np.zeros((self.th, self.tw, 4), dtype=np.uint8)

                tiff_tile_row = int(np.ceil(ttile_num // num_tiff_tile_cols))
                tiff_tile_col = int(np.ceil(ttile_num % num_tiff_tile_cols))
                x = tiff_tile_col * self.tw
                y = tiff_tile_row * self.th

                libtiff.readRGBATile(self.tiff_fp, x, y, rgba_tile)

                # The RGBA interface requires some reordering.
                if sys.byteorder == "little":
                    # image is upside down
                    dims = [0]
                else:
                    # image is upside down, but in addition,
                    # if big-endian, must also flip the image planes
                    dims = [0, 2]
                rgba_tile = np.flip(rgba_tile, axis=dims)

                # We may need to remove the alpha plane.
                tiff_tile = rgba_tile[:, :, :3]

            else:

                tiff_tile = np.zeros(
                    (self.th, self.tw, self.spp),
                    dtype=self.dtype
                )
                libtiff.readEncodedTile(self.tiff_fp, ttile_num, tiff_tile)

            # determine how to fit this tiff tile into the jp2k
            # tile
            #
            # these are the section coordinates in image space
            uly = max(jp2k_uly, tile_uly)
            lly = min(jp2k_uly + jth, tile_uly + self.th)

            ulx = max(jp2k_ulx, tile_ulx)
            urx = min(jp2k_ulx + jtw, tile_ulx + self.tw)

            # convert to JP2K tile coordinates
            jrows = slice(uly % jth, (lly - 1) % jth + 1)
            jcols = slice(ulx % jtw, (urx - 1) % jtw + 1)

            # convert to TIFF tile coordinates
            trows = slice(uly % self.th, (lly - 1) % self.th + 1)
            tcols = slice(ulx % self.tw, (urx - 1) % self.tw + 1)

            jp2k_tile[jrows, jcols, :] = tiff_tile[trows, tcols, :]

        # last tile column?  last tile row?  If so, we may have a partial tile.
        if partial_jp2_tile_cols and jp2k_tile_col == num_jp2k_tile_cols - 1:
            last_j2k_cols = slice(0, self.imagewidth - jp2k_ulx)
            jp2k_tile = jp2k_tile[:, last_j2k_cols, :].copy()

        if partial_jp2_tile_rows and jp2k_tile_row == num_jp2k_tile_rows - 1:
            last_j2k_rows = slice(0, self.imageheight - jp2k_uly)
            jp2k_tile = jp2k_tile[last_j2k_rows, :, :].copy()

        return jp2k_tile

    def _get_covering_tiles(self, jp2k_tile_num):
        """
        Construct the set of TIFF tiles that completely cover the jpeg2000
        tile.
        """
        jth, jtw = self.tilesize

        num_jp2k_tile_cols = int(np.ceil(self.imagewidth / jtw))

        jp2k_tile_row = int(np.ceil(jp2k_tile_num // num_jp2k_tile_cols))
        jp2k_tile_col = int(np.ceil(jp2k_tile_num % num_jp2k_tile_cols))

        num_tiff_tile_cols = int(np.ceil(self.imagewidth / self.tw))

        # Upper left corner of the jp2k tile
        ulx = jtw * jp2k_tile_col
        uly = jth * jp2k_tile_row
        ul_tiff_tilenum = libtiff.computeTile(self.tiff_fp, ulx, uly, 0, 0)
        left_tiff_tile_col = int(np.ceil(ulx // self.tw))

        # Upper right corner of the jp2k tile
        urx = min(ulx + jtw - 1, self.imagewidth - 1)
        right_tiff_tile_col = int(np.ceil(urx // self.tw))

        # lower left corner
        llx = ulx
        lly = min(uly + jth - 1, self.imageheight - 1)

        # lower right corner
        lrx = min(llx + jtw - 1, self.imagewidth - 1)
        lry = lly
        lr_tiff_tilenum = libtiff.computeTile(self.tiff_fp, lrx, lry, 0, 0)

        # collect the tiles
        tiles = set()

        for tile_num in range(ul_tiff_tilenum, lr_tiff_tilenum + 1):

            tiff_tile_col = int(np.ceil(tile_num % num_tiff_tile_cols))

            # The tile rows should always be good.  But we need to exclude
            # any tiles based on the tile column.
            if (
                tiff_tile_col < left_tiff_tile_col
                or tiff_tile_col > right_tiff_tile_col
            ):
                continue

            # otherwise, this is a tile that intersects the jp2k tile
            tiles.add(tile_num)

        return tiles

    def _write_striped_tiff_to_tiled_jp2k(self):
        """The input TIFF image is striped and we are to create the output
        JPEG2000 image as a tiled JP2K.
        """

        jth, jtw = self.tilesize

        self.logger.debug(f"image:  {self.imageheight} x {self.imagewidth}")
        self.logger.debug(f"jptile:  {jth} x {jtw}")
        num_strips = libtiff.numberOfStrips(self.tiff_fp)

        num_jp2k_tile_cols = int(np.ceil(self.imagewidth / jtw))

        for idx, tilewriter in enumerate(self.jp2.get_tilewriters()):

            jp2k_tile_row = idx // num_jp2k_tile_cols
            jp2k_tile_col = idx % num_jp2k_tile_cols

            msg = f"Tile:  #{idx} row #{jp2k_tile_row} col #{jp2k_tile_col}"
            self.logger.info(msg)

            # the coordinates of the upper left pixel of the jp2k tile
            july, julx = jp2k_tile_row * jth, jp2k_tile_col * jtw

            # If we are starting a new row of jp2k tiles, we want to allocate
            # space for all the TIFF strips that encompass this jp2k row, and
            # then go ahead and read them in.  For all other cases, just assign
            # jp2k tiles from this same TIFF multi-strip.
            if jp2k_tile_col == 0:
                tiff_multi_strip = self._construct_multi_strip(
                    july,
                    num_strips,
                    jth,
                )

            # construct the TIFF row and column slices from the multi-strip,
            # assign to the jp2k tile.
            stripnum = libtiff.computeStrip(self.tiff_fp, july, 0)
            tile_uly = stripnum * self.rps

            ms_uly = july - tile_uly
            ms_ulx = julx

            ms_lry = min(ms_uly + jth, self.imageheight - tile_uly)
            ms_lrx = min(ms_ulx + jtw, self.imagewidth)

            rows = slice(ms_uly, ms_lry)
            cols = slice(ms_ulx, ms_lrx)

            tilewriter[:] = tiff_multi_strip[rows, cols, :]

    def _construct_multi_strip(self, july, num_strips, jth):
        """TIFF strips are pretty inefficient.  If our I/O was stupidly focused
        solely on each JP2K tile, we would read in each TIFF strip multiple
        times, once for each JP2K tile in the JP2K tile row.  If instead, we
        read in ALL the strips that will encompass that current row of JP2K
        tiles, we will save ourselves a lot of disk I/O.

        Parameters
        ----------
        jth : int
            The number of rows in a JP2K tile.
        july : int
            The top row of the current JP2K tile row.

        Returns
        -------
        tiff_multi_strip : np.array
            Holds all the TIFF strips that tightly encompass the current JP2K
            tile row.
        """
        num_strips = libtiff.numberOfStrips(self.tiff_fp)

        # We need to create a TIFF "multi-strip" that can hold all of
        # the JP2K tiles in a JP2K tile row.
        y = july
        top_strip_num = libtiff.computeStrip(self.tiff_fp, y, 0)

        # Find the strip number that is ONE MORE than the bottom strip.
        while (y // self.rps) * self.rps < min(july + jth, self.imageheight):
            y += self.rps
        bottom_strip_num = libtiff.computeStrip(self.tiff_fp, y, 0)

        # compute the number of rows contained between the top strip
        # and the bottom strip
        num_rows = (bottom_strip_num - top_strip_num) * self.rps

        if self.photo == libtiff.Photometric.YCBCR:
            # always single byte samples of R, G, B, and A
            dtype = np.uint8
        else:
            dtype = self.dtype
        spp = self.spp

        # This may result in a multi-strip that has more rows than the jp2k
        # tile
        tiff_multi_strip = np.zeros(
            (num_rows, self.imagewidth, spp),
            dtype=dtype
        )

        # Fill the multi-strip
        for stripnum in range(top_strip_num, bottom_strip_num):

            if self.photo == libtiff.Photometric.YCBCR:

                tiff_rgba_strip = np.zeros(
                    (self.rps, self.imagewidth, 4),
                    dtype=dtype
                )

                libtiff.readRGBAStrip(
                    self.tiff_fp, stripnum * self.rps, tiff_rgba_strip
                )

                # If a partial last strip...
                if (
                    stripnum == num_strips - 1
                    and self.imageheight // self.rps != num_strips
                ):
                    # According to the man page:
                    #
                    # When reading a partial last strip in the file the last
                    # line of the image  will  begin at the beginning of the
                    # buffer.
                    #
                    # move the top strips down to the bottom, otherwise the
                    # following flipping logic doesn't work.
                    bottom_row = self.imageheight % self.rps
                    irows = slice(0, bottom_row)
                    orows = slice(self.rps - bottom_row, self.rps)
                    tiff_rgba_strip[orows, :, :] = tiff_rgba_strip[irows, :, :]

                # The rgba interface requires at least flipping the image
                # upside down, and also reordering the planes on big endian
                if sys.byteorder == "little":
                    dims = [0]
                else:
                    dims = [0, 2]
                tiff_rgba_strip = np.flip(tiff_rgba_strip, axis=dims)

                # potentially get rid of alpha plane
                tiff_strip = tiff_rgba_strip[:, :, : self.spp]

            else:

                tiff_strip = np.zeros(
                    (self.rps, self.imagewidth, spp),
                    dtype=dtype
                )

                libtiff.readEncodedStrip(self.tiff_fp, stripnum, tiff_strip)

            # push the strip into the multi-strip
            top_row = (stripnum - top_strip_num) * self.rps
            bottom_row = (stripnum - top_strip_num + 1) * self.rps
            rows = slice(top_row, bottom_row)

            tiff_multi_strip[rows, :, :] = tiff_strip

        return tiff_multi_strip



# standard library imports
import logging
import shutil
import unittest
import uuid
import warnings

# 3rd party library imports
import numpy as np
import skimage

# Local imports
import glymur
from glymur import Jp2k, JPEG2JP2
from glymur.core import SRGB
from . import fixtures
from .fixtures import (
    OPENJPEG_NOT_AVAILABLE,
    OPENJPEG_NOT_AVAILABLE_MSG,
    CANNOT_USE_IMPORTLIB_METADATA
)


@unittest.skipIf(OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG)
@unittest.skipIf(
    CANNOT_USE_IMPORTLIB_METADATA,
    'missing importlib.metadata.files ?'
)
class TestSuite(fixtures.TestJPEGCommon):

    def test_smoke(self):
        """
        SCENARIO:  Convert JPEG without metadata to JP2

        EXPECTED RESULT:  data matches, it's just one big tile
        """

        with JPEG2JP2(self.retina, self.temp_jp2_filename) as p:
            p.run()

        j = Jp2k(self.temp_jp2_filename)

        actual = j[:]
        expected = skimage.data.retina()

        # data matches
        np.testing.assert_array_equal(actual, expected)

        [h, w, _] = actual.shape

        # it's one big tile
        c = j.get_codestream()
        self.assertEqual(c.segment[1].xtsiz, w)
        self.assertEqual(c.segment[1].ytsiz, h)

    def test_default_action_icc_profile(self):
        """
        SCENARIO:  Convert JPEG with an ICC profile using default options

        EXPECTED RESULT:  data matches, it's just one big tile.  The JP2 has
        no ICC profile information.  There is a warning about a skipped ICC
        profile.
        """

        with (
            JPEG2JP2(self.rocket, self.temp_jp2_filename) as p,
            self.assertLogs(logger='tiff2jp2', level=logging.INFO) as cm,
        ):
            p.run()

        self.assertEqual(sum('ICC profile' in msg for msg in cm.output), 1)

        j = Jp2k(self.temp_jp2_filename)

        actual = j[:]
        expected = skimage.data.rocket()

        # data matches
        np.testing.assert_array_equal(actual, expected)

        # The colour specification box does not have the profile
        colr = j.box[2].box[1]
        self.assertEqual(colr.method, glymur.core.ENUMERATED_COLORSPACE)
        self.assertEqual(colr.precedence, 0)
        self.assertEqual(colr.approximation, 0)
        self.assertEqual(colr.colorspace, SRGB)
        self.assertIsNone(colr.icc_profile)

    def test_embed_icc_profile(self):
        """
        SCENARIO:  Convert JPEG with an ICC profile, specifically setting the
        include_icc_profile keyword.

        EXPECTED RESULT:  The JP2 has ICC profile information.
        """

        with (
            JPEG2JP2(
                self.rocket,
                self.temp_jp2_filename,
                include_icc_profile=True
            ) as p,
        ):
            p.run()

        j = Jp2k(self.temp_jp2_filename)

        actual = j[:]
        expected = skimage.data.rocket()

        # data matches
        np.testing.assert_array_equal(actual, expected)

        # The colour specification box has the profile
        self.assertIsNotNone(j.box[2].box[1].icc_profile)

    def test_tilesize(self):
        """
        SCENARIO:  Convert JPEG without metadata to JP2, specify tile size

        EXPECTED RESULT:  data matches, tile size matches
        """

        kwargs = {'tilesize': (512, 512)}
        with JPEG2JP2(self.retina, self.temp_jp2_filename, **kwargs) as j:
            j.run()

        jp2 = Jp2k(self.temp_jp2_filename)

        actual = jp2[:]
        expected = skimage.data.retina()

        np.testing.assert_array_equal(actual, expected)

        c = jp2.get_codestream()
        self.assertEqual(c.segment[1].xtsiz, 512)
        self.assertEqual(c.segment[1].ytsiz, 512)

    def test_exif_xmp(self):
        """
        SCENARIO:  Convert JPEG with EXIF and XMP metadata to JP2

        EXPECTED RESULT:  data matches, there is an EXIF UUID box, there is an
        XMP UUID box
        """

        with JPEG2JP2(self.hubble, self.temp_jp2_filename) as p:
            p.run()

        j = Jp2k(self.temp_jp2_filename)

        actual = j[:]
        expected = skimage.data.hubble_deep_field()

        np.testing.assert_array_equal(actual, expected)

        def exif_predicate(x):
            exif_uuid = uuid.UUID(bytes=b"JpgTiffExif->JP2")
            if x.box_id == 'uuid' and x.uuid == exif_uuid:
                return True
            else:
                return False

        box = next(filter(exif_predicate, j.box), None)
        self.assertIsNotNone(box)

        def xmp_predicate(x):
            xmp_uuid = uuid.UUID("be7acfcb-97a9-42e8-9c71-999491e3afac")
            if x.box_id == 'uuid' and x.uuid == xmp_uuid:
                return True
            else:
                return False

        box = next(filter(xmp_predicate, j.box), None)
        self.assertIsNotNone(box)

    def test_verbosity(self):
        """
        SCENARIO:  Convert JPEG to JP2, use WARN log level.

        EXPECTED RESULT:  data matches, one message detected at INFO level
        """
        with (
            JPEG2JP2(
                self.retina, self.temp_jp2_filename, verbosity=logging.INFO
            ) as p,
            self.assertLogs(logger='tiff2jp2', level=logging.INFO) as cm
        ):
            p.run()

            self.assertEqual(len(cm.output), 1)

    def test_psnr(self):
        """
        SCENARIO:  Convert JPEG file to JP2 with the psnr keyword argument

        EXPECTED RESULT:  data matches
        """
        with JPEG2JP2(
            self.retina, self.temp_jp2_filename, psnr=(30, 35, 40, 0)
        ) as p:
            p.run()

        j = Jp2k(self.temp_jp2_filename)

        d = {}
        for layer in range(4):
            j.layer = layer
            d[layer] = j[:]

        truth = skimage.io.imread(self.retina)

        with warnings.catch_warnings():
            # MSE is zero for that first image, resulting in a divide-by-zero
            # warning
            warnings.simplefilter('ignore')
            psnr = [
                skimage.metrics.peak_signal_noise_ratio(truth, d[j])
                for j in range(4)
            ]

        # That first image should be lossless.
        self.assertTrue(np.isinf(psnr[0]))

        # None of the subsequent images should have inf PSNR.
        self.assertTrue(not np.any(np.isinf(psnr[1:])))

        # PSNR should increase for the remaining images.
        self.assertTrue(np.all(np.diff(psnr[1:])) > 0)

    def test_existing_file(self):
        """
        Scenario:  provide an existing JP2 file as the output file

        Expected Result:  RuntimeError
        """
        shutil.copyfile(glymur.data.nemo(), self.temp_jp2_filename)
        with (
            self.assertRaises(FileExistsError),
            JPEG2JP2(self.retina, self.temp_jp2_filename) as p,
        ):
            p.run()



"""
Test suite for warnings issued by glymur.
"""
# Standard library imports
import codecs
import importlib.resources as ir
from io import BytesIO
import struct
import unittest
import warnings

# 3rd party library imports
import numpy as np

# Local imports
from glymur import Jp2k
import glymur
from glymur.core import COLOR, RED, GREEN, BLUE
from glymur.jp2box import InvalidJp2kError

from . import fixtures
from .fixtures import OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG


class TestSuite(fixtures.TestCommon):

    def setUp(self):
        super().setUp()

        # Reset printoptions for every test.
        glymur.reset_option('all')

    def tearDown(self):
        super().tearDown()

        warnings.resetwarnings()
        glymur.reset_option('all')

    def test_parsing_bad_fptr_box(self):
        """
        SCENARIO: An ftyp box advertises too many bytes to be read.

        EXPECTED RESULT:  A warning is issued.  In this case we also end up
        erroring out anyway since we don't get a valid FileType box.
        """
        path = ir.files('tests.data').joinpath('issue438.jp2')
        with self.assertWarns(UserWarning):
            with self.assertRaises(InvalidJp2kError):
                Jp2k(path)

    def test_siz_ihdr_mismatch(self):
        """
        SCENARIO:  The dimensions reported by the IHDR box don't match what is
        reported by the SIZ marker.

        EXPECTED RESULT: A warning is issued.
        """
        with open(self.temp_j2k_filename, mode='wb') as tfile:
            with open(self.jp2file, 'rb') as ifile:
                # Everything up until the IHDR payload
                read_buffer = ifile.read(48)
                tfile.write(read_buffer)

                # Write the bad IHDR.  The correct sequence of values read
                # should be
                # (1456, 2592, 3, 7, 7, 0, 0)
                bad_ihdr = (1600, 2592, 3, 7, 7, 0, 0)
                buffer = struct.pack('>IIHBBBB', *bad_ihdr)
                tfile.write(buffer)

                # Get the rest of the input file.
                ifile.seek(62)
                read_buffer = ifile.read()
                tfile.write(read_buffer)
                tfile.flush()

            with self.assertWarns(UserWarning):
                # c = Jp2k(tfile.name).get_codestream(header_only=False)
                Jp2k(tfile.name)

    def test_unrecoverable_xml(self):
        """
        Bad byte sequence in XML that cannot be parsed.

        Original test file was
        26ccf3651020967f7778238ef5af08af.SIGFPE.d25.527.jp2
        """
        fptr = BytesIO()

        payload = b'\xees'
        fptr.write(payload)
        fptr.seek(0)

        with self.assertWarns(UserWarning):
            box = glymur.jp2box.XMLBox.parse(fptr, 0, 8 + len(payload))

        self.assertIsNone(box.xml)

    def test_invalid_progression_order(self):
        """
        Should still be able to parse even if prog order is invalid.

        Original test file was input/nonregression/2977.pdf.asan.67.2198.jp2
        """
        fp = BytesIO()
        buffer = struct.pack('>HBBBBBBBBBB', 12, 3, 33, 1, 1, 3, 3, 0, 0, 1, 1)
        fp.write(buffer)
        fp.seek(0)

        with self.assertWarns(UserWarning):
            glymur.codestream.Codestream._parse_cod_segment(fp)

    def test_bad_wavelet_transform(self):
        """
        Should warn if wavelet transform is bad.  Issue195

        Original file tested was input/nonregression/edf_c2_10025.jp2
        """
        fp = BytesIO()
        buffer = struct.pack('>HBHBBBBBBB', 12, 0, 1, 1, 1, 3, 3, 0, 0, 10)
        fp.write(buffer)
        fp.seek(0)

        with self.assertWarns(UserWarning):
            glymur.codestream.Codestream._parse_cod_segment(fp)

    def test_NR_gdal_fuzzer_assert_in_opj_j2k_read_SQcd_SQcc_patch_jp2(self):
        """
        validate the QCC component number against Csiz

        The original test file was
        gdal_fuzzer_assert_in_opj_j2k_read_SQcd_SQcc.patch.jp2
        """
        fp = BytesIO()

        buffer = struct.pack('>HBB', 4, 64, 64)
        fp.write(buffer)
        fp.seek(0)

        with self.assertWarns(UserWarning):
            glymur.codestream.Codestream._parse_qcc_segment(fp)

    def test_NR_gdal_fuzzer_check_comp_dx_dy_jp2_dump(self):
        """
        Invalid subsampling value.

        Original test file was gdal_fuzzer_check_comp_dx_dy.jp2
        """
        fp = BytesIO()

        buffer = struct.pack('>H', 47)  # length

        # kwargs = {'rsiz': 1,
        #           'xysiz': (1000, 1000),
        #           'xyosiz': (0, 0),
        #           'xytsiz': (1000, 1000),
        #           'xytosiz': (0, 0),
        #           'Csiz': 3,
        #           'bitdepth': (8, 8, 8),
        #           'signed':  (False, False, False),
        #           'xyrsiz': ((1, 1, 1), (1, 1, 1)),
        #           'length': 47,
        #           'offset': 2}
        buffer += struct.pack('>HIIIIIIIIH', 1, 1000, 1000, 0, 0, 1000, 1000,
                              0, 0, 3)
        buffer += struct.pack('>BBBBBBBBB', 7, 1, 1, 7, 1, 1, 7, 1, 0)
        fp.write(buffer)
        fp.seek(0)

        with self.assertWarns(UserWarning):
            glymur.codestream.Codestream._parse_siz_segment(fp)

    def test_read_past_end_of_box(self):
        """
        SCENARIO:  A pclr box has more rows specified than can fit inside the
        given box length.

        EXPECTED RESULT:  A warning is issued for an attempt to read past the
        end of the box.
        """
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jpxfile, 'rb') as ifile:
                ofile.write(ifile.read(93))

                # Rewrite the ncols, nrows portion.  Increase the number of
                # rows.  This causes python to think there are more rows
                # than there actually are when resizing the palette.
                buffer = struct.pack('>HB', 257, 3)
                ofile.write(buffer)

                ifile.seek(96)
                ofile.write(ifile.read())
                ofile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(ofile.name)

    def test_NR_gdal_fuzzer_check_number_of_tiles(self):
        """
        Has an impossible tiling setup.

        Original test file was input/nonregression
                               /gdal_fuzzer_check_number_of_tiles.jp2
        """
        fp = BytesIO()

        buffer = struct.pack('>H', 47)  # length

        # kwargs = {'rsiz': 1,
        #           'xysiz': (20, 16777236),
        #           'xyosiz': (0, 0),
        #           'xytsiz': (20, 20),
        #           'xytosiz': (0, 0),
        #           'Csiz': 3,
        #           'bitdepth': (8, 8, 8),
        #           'signed':  (False, False, False),
        #           'xyrsiz': ((1, 1, 1), (1, 1, 1)),
        #           'length': 47,
        #           'offset': 2}
        buffer += struct.pack('>HIIIIIIIIH', 1, 20, 16777236, 0, 0, 20, 20,
                              0, 0, 3)
        buffer += struct.pack('>BBBBBBBBB', 7, 1, 1, 7, 1, 1, 7, 1, 1)
        fp.write(buffer)
        fp.seek(0)

        with self.assertWarns(UserWarning):
            glymur.codestream.Codestream._parse_siz_segment(fp)

    def test_NR_gdal_fuzzer_unchecked_numresolutions_dump(self):
        """
        Has an invalid number of resolutions.

        Original test file was input/nonregression/
                               gdal_fuzzer_unchecked_numresolutions.jp2
        """
        pargs = (0, 0, 1, 1, 64, 3, 3, 0, 0, None)
        spcod = struct.pack('>BHBBBBBB', 0, 1, 1, 64, 3, 3, 0, 0)
        spcod = bytearray(spcod)
        with self.assertWarns(UserWarning):
            glymur.codestream.CODsegment(*pargs, length=12, offset=174)

    def test_file_pointer_badly_positioned(self):
        """
        SCENARIO:  A colr box has an impossibly too long box length.   Since
        the colr box is the last one in the jp2h super box, this results in
        an attempt to read past the end of the super box.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jp2file, 'rb') as ifile:

                # Write up to the colr box
                ofile.write(ifile.read(62))

                # Write a too-long color box
                buffer = struct.pack('>I4sBBBI',
                                     4194319, b'colr', 1, 0, 0, 0)
                ofile.write(buffer)

                # Write everything past the colr box.
                ifile.seek(77)
                ofile.write(ifile.read())
                ofile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(ofile.name)

    def test_NR_DEC_issue188_beach_64bitsbox_jp2_41_decode(self):
        """
        SCENARIO:  A JP2 file has a box with label 'XML ' instead of 'xml '.

        EXPECTED RESULT:  A warning is issued about the box being unrecognized.
        """
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jp2file, 'rb') as ifile:
                ofile.write(ifile.read())

                buffer = struct.pack('>I4s', 32, b'XML ')
                s = "<stuff>goes here</stuff>"
                buffer += s.encode('utf-8')
                ofile.write(buffer)
                ofile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(ofile.name)

    def test_truncated_icc_profile(self):
        """
        Validate a warning for a truncated ICC profile
        """
        obj = BytesIO()
        obj.write(b'\x00' * 66)

        # Write a colr box with a truncated ICC profile.
        # profile.
        buffer = struct.pack('>I4s', 47, b'colr')
        buffer += struct.pack('>BBB', 2, 0, 0)

        buffer += b'\x00' * 12 + b'scnr' + b'XYZ ' + b'Lab '
        # Need a date in bytes 24:36
        buffer += struct.pack('>HHHHHH', 1966, 2, 15, 0, 0, 0)
        obj.write(buffer)
        obj.seek(74)

        # Should be able to read the colr box now
        with self.assertWarns(UserWarning):
            glymur.jp2box.ColourSpecificationBox.parse(obj, 66, 47)

    def test_invalid_colour_specification_method(self):
        """
        should not error out with invalid colour specification method
        """
        obj = BytesIO()
        obj.write(b'\x00' * 66)

        # Write a colr box with a bad method (254).  This requires an ICC
        # profile.
        buffer = struct.pack('>I4s', 143, b'colr')
        buffer += struct.pack('>BBB', 254, 0, 0)

        buffer += b'\x00' * 12 + b'scnr' + b'XYZ ' + b'Lab '
        # Need a date in bytes 24:36
        buffer += struct.pack('>HHHHHH', 1966, 2, 15, 0, 0, 0)
        buffer += b'\x00' * 92
        obj.write(buffer)
        obj.seek(74)

        # Should be able to read the colr box now
        with self.assertWarns(UserWarning):
            glymur.jp2box.ColourSpecificationBox.parse(obj, 66, 143)

    def test_bad_color_space_specification(self):
        """
        Verify that a warning is issued if the color space method is invalid.

        For JP2, the method must be either 1 or 2.
        """
        jp2 = glymur.Jp2k(self.jp2file)
        jp2.box[2].box[1].method = 3
        with self.assertWarns(UserWarning):
            jp2._validate()

    def test_unknown_superbox(self):
        """
        SCENARIO:  There is a superbox with an unrecognized label.

        EXPECTED RESULT:  A warning is issued.
        """

        with open(self.temp_jpx_filename, mode='wb') as tfile:
            with open(self.jpxfile, 'rb') as ifile:
                tfile.write(ifile.read())

            # Add the header for an unknown superbox.
            write_buffer = struct.pack('>I4s', 20, 'grp '.encode())
            tfile.write(write_buffer)

            # Add a free box inside of it.  We won't be able to identify it,
            # but it's there.
            write_buffer = struct.pack('>I4sI', 12, 'free'.encode(), 0)
            tfile.write(write_buffer)
            tfile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(tfile.name)

    def test_brand_unknown(self):
        """A ftyp box brand must be 'jp2 ' or 'jpx '."""
        with self.assertWarns(UserWarning):
            glymur.jp2box.FileTypeBox(brand='jp3')

    def test_bad_type(self):
        """Channel types are limited to 0, 1, 2, 65535
        Should reject if not all of index, channel_type, association the
        same length.
        """
        channel_type = (COLOR, COLOR, 3)
        association = (RED, GREEN, BLUE)
        with self.assertWarns(UserWarning):
            glymur.jp2box.ChannelDefinitionBox(channel_type=channel_type,
                                               association=association)

    def test_wrong_lengths(self):
        """Should reject if not all of index, channel_type, association the
        same length.
        """
        channel_type = (COLOR, COLOR)
        association = (RED, GREEN, BLUE)
        with self.assertWarns(UserWarning):
            glymur.jp2box.ChannelDefinitionBox(channel_type=channel_type,
                                               association=association)

    def test_cl_entry_unknown(self):
        """A ftyp box cl list can only contain 'jp2 ', 'jpx ', or 'jpxb'."""
        with self.assertWarns(UserWarning):
            # Bad compatibility list item.
            glymur.jp2box.FileTypeBox(compatibility_list=['jp3'])

    def test_colr_with_cspace_and_icc(self):
        """Colour specification boxes can't have both."""
        buffer = ir.files('tests.data').joinpath('sgray.icc').read_bytes()

        with self.assertWarns(UserWarning):
            colorspace = glymur.core.SRGB
            glymur.jp2box.ColourSpecificationBox(colorspace=colorspace,
                                                 icc_profile=buffer)

    def test_colr_with_bad_method(self):
        """colr must have a valid method field"""
        colorspace = glymur.core.SRGB
        method = -1
        with self.assertWarns(UserWarning):
            glymur.jp2box.ColourSpecificationBox(colorspace=colorspace,
                                                 method=method)

    def test_colr_with_bad_approx(self):
        """
        SCENARIO:  An ColourSpecificationBox is given an invalid approximation
        value.

        EXPECTED RESULT:  A warning is issued.
        """
        with self.assertWarns(UserWarning):
            glymur.jp2box.ColourSpecificationBox(colorspace=glymur.core.SRGB,
                                                 approximation=-1)

    def test_mismatched_bitdepth_signed(self):
        """bitdepth and signed arguments must have equal length"""
        palette = np.array([[255, 0, 255], [0, 255, 0]], dtype=np.uint8)
        bps = (8, 8, 8)
        signed = (False, False)
        with self.assertWarns(UserWarning):
            glymur.jp2box.PaletteBox(palette, bits_per_component=bps,
                                     signed=signed)

    def test_mismatched_signed_palette(self):
        """bitdepth and signed arguments must have equal length"""
        palette = np.array([[255, 0, 255], [0, 255, 0]], dtype=np.uint8)
        bps = (8, 8, 8, 8)
        signed = (False, False, False, False)
        with self.assertWarns(UserWarning):
            glymur.jp2box.PaletteBox(palette, bits_per_component=bps,
                                     signed=signed)

    def test_invalid_xml_box(self):
        """
        SCENARIO:  An xml box has invalid XML.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as tfile:
            bad_xml_file = tfile.name
            with open(self.jp2file, 'rb') as ifile:
                # Everything up until the UUID box.
                write_buffer = ifile.read(77)
                tfile.write(write_buffer)

                # Write the xml box with bad xml
                # Length = 28, id is 'xml '.
                write_buffer = struct.pack('>I4s', int(28), b'xml ')
                tfile.write(write_buffer)

                write_buffer = '<test>this is a test'
                write_buffer = write_buffer.encode()
                tfile.write(write_buffer)

                # Get the rest of the input file.
                write_buffer = ifile.read()
                tfile.write(write_buffer)
                tfile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(bad_xml_file)

    def test_deurl_child_of_dtbl(self):
        """
        Data reference boxes can only contain data entry url boxes.

        It's just a warning here because we haven't tried to write it.
        """
        ftyp = glymur.jp2box.FileTypeBox()
        with self.assertWarns(UserWarning):
            glymur.jp2box.DataReferenceBox([ftyp])

    def test_flst_lens_not_the_same(self):
        """A fragment list box items must be the same length."""
        offset = [89]
        length = [1132288]
        reference = [0, 0]
        with self.assertWarns(UserWarning):
            glymur.jp2box.FragmentListBox(offset, length, reference)

    def test_flst_offsets_not_positive(self):
        """A fragment list box offsets must be positive."""
        offset = [0]
        length = [1132288]
        reference = [0]
        with self.assertWarns(UserWarning):
            glymur.jp2box.FragmentListBox(offset, length, reference)

    def test_flst_lengths_not_positive(self):
        """A fragment list box lengths must be positive."""
        offset = [89]
        length = [0]
        reference = [0]
        with self.assertWarns(UserWarning):
            glymur.jp2box.FragmentListBox(offset, length, reference)

    def test_unrecognized_exif_tag(self):
        """
        SCENARIO:  An Exif UUID box has an unrecognized tag.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as tfile:

            with open(self.jp2file, 'rb') as ifptr:
                tfile.write(ifptr.read())

            # Write L, T, UUID identifier.
            tfile.write(struct.pack('>I4s', 52, b'uuid'))
            tfile.write(b'JpgTiffExif->JP2')

            tfile.write(b'Exif\x00\x00')
            xbuffer = struct.pack('<BBHI', 73, 73, 42, 8)
            tfile.write(xbuffer)

            # We will write just a single tag.
            tfile.write(struct.pack('<H', 1))

            # The "Make" tag is tag no. 271.  Corrupt it to 171.
            tfile.write(struct.pack('<HHI4s', 171, 2, 3, b'HTC\x00'))
            tfile.flush()

            with self.assertWarns(UserWarning):
                glymur.Jp2k(tfile.name)

    def test_bad_tag_datatype(self):
        """
        SCENARIO:  A tag with an unrecognized numeric datatype field is found
        in an Exif UUID box.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as tfile:

            with open(self.jp2file, 'rb') as ifptr:
                tfile.write(ifptr.read())

            # Write L, T, UUID identifier.
            tfile.write(struct.pack('>I4s', 52, b'uuid'))
            tfile.write(b'JpgTiffExif->JP2')

            tfile.write(b'Exif\x00\x00')
            xbuffer = struct.pack('<BBHI', 73, 73, 42, 8)
            tfile.write(xbuffer)

            # We will write just a single tag.
            tfile.write(struct.pack('<H', 1))

            # 2000 is not an allowable TIFF datatype.
            tfile.write(struct.pack('<HHI4s', 271, 2000, 3, b'HTC\x00'))
            tfile.flush()

            with self.assertWarns(UserWarning):
                glymur.Jp2k(tfile.name)

    def test_bad_tiff_header_byte_order_indication(self):
        """
        SCENARIO:  An invalid TIFF header byte order marker is encountered
        in an Exif UUID box.  Only b'II' and b'MM' are allowed.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as tfile:

            with open(self.jp2file, 'rb') as ifptr:
                tfile.write(ifptr.read())

            # Write L, T, UUID identifier.
            tfile.write(struct.pack('>I4s', 52, b'uuid'))
            tfile.write(b'JpgTiffExif->JP2')

            tfile.write(b'Exif\x00\x00')

            # Here's the bad byte order.
            tfile.write(b'JI')

            # Write the rest of the header.
            xbuffer = struct.pack('<HI', 42, 8)
            tfile.write(xbuffer)

            # We will write just a single tag.
            tfile.write(struct.pack('<H', 1))

            # 271 is the Make.
            tfile.write(struct.pack('<HHI4s', 271, 2, 3, b'HTC\x00'))
            tfile.flush()

            with self.assertWarns(UserWarning):
                glymur.Jp2k(tfile.name)

    @unittest.skipIf(OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG)
    def test_warn_if_using_read_method(self):
        """Should warn if deprecated read method is called"""
        with self.assertWarns(DeprecationWarning):
            Jp2k(self.jp2file).read()

    def test_bad_rsiz(self):
        """
        SCENARIO:  The SIZ value parsed from the SIZ segment is invalid.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jp2file, 'rb') as ifile:
                # Copy up until the RSIZ value.
                ofile.write(ifile.read(91))

                # Write the bad RSIZ value.
                buffer = struct.pack('>H', 32)
                ofile.write(buffer)
                ifile.seek(93)

                # Get the rest of the file.
                ofile.write(ifile.read())

                ofile.seek(0)

            with self.assertWarns(UserWarning):
                Jp2k(ofile.name)

    def test_undecodeable_box_id(self):
        """
        SCENARIO:  an unknown box ID is encountered

        EXPECTED RESULT:  Should warn but not error out.
        """
        bad_box_id = b'abcd'
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jp2file, 'rb') as ifile:
                ofile.write(ifile.read())

                # Tack an unrecognized box onto the end of nemo.
                buffer = struct.pack('>I4s', 8, bad_box_id)
                ofile.write(buffer)
                ofile.flush()

            with self.assertWarns(UserWarning):
                jp2 = Jp2k(ofile.name)

            # Now make sure we got all of the boxes.
            box_ids = [box.box_id for box in jp2.box]
            self.assertEqual(box_ids, ['jP  ', 'ftyp', 'jp2h', 'jp2c', 'xxxx'])
            self.assertEqual(jp2.box[-1].claimed_box_id, b'abcd')

    def test_bad_ftyp_brand(self):
        """
        SCENARIO:  The ftyp box has an invalid brand field.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jp2file, 'rb') as ifile:
                # Write the JPEG2000 signature box
                ofile.write(ifile.read(12))

                # Write a bad version of the file type box.  'jp  ' is not
                # allowed as a brand.
                buffer = struct.pack('>I4s4sI4s', 20, b'ftyp', b'jp  ', 0,
                                     b'jp2 ')
                ofile.write(buffer)

                # Write the rest of the boxes as-is.
                ifile.seek(32)
                ofile.write(ifile.read())
                ofile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(ofile.name)

    def test_bad_ftyp_compatibility_list_item(self):
        """
        SCENARIO:  The ftyp box has an invalid compatibility list item.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jp2file, 'rb') as ifile:
                # Write the JPEG2000 signature box
                ofile.write(ifile.read(12))

                # Write a bad compatibility list item.  'jp3' is not valid.
                buffer = struct.pack('>I4s4sI4s', 20, b'ftyp', b'jp2 ', 0,
                                     b'jp3 ')
                ofile.write(buffer)

                # Write the rest of the boxes as-is.
                ifile.seek(32)
                ofile.write(ifile.read())
                ofile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(ofile.name)

    def test_invalid_approximation(self):
        """
        SCENARIO:  The colr box has an invalid approximation field.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jp2file, 'rb') as ifile:
                # Copy the signature, file type, and jp2 header, image header
                # box as-is.
                ofile.write(ifile.read(62))

                # Write a bad version of the color specification box.  32 is an
                # invalid approximation value.
                buffer = struct.pack('>I4sBBBI', 15, b'colr', 1, 2, 32, 16)
                ofile.write(buffer)

                # Write the rest of the boxes as-is.
                ifile.seek(77)
                ofile.write(ifile.read())
                ofile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(ofile.name)

    def test_invalid_colorspace(self):
        """
        SCENARIO:  A colr box has an invalid colorspace field.

        EXPECTED RESULT:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jp2file, 'rb') as ifile:
                # Copy the signature, file type, and jp2 header, image header
                # box as-is.
                ofile.write(ifile.read(62))

                # Write a bad version of the color specification box.  276 is
                # an invalid colorspace.
                buffer = struct.pack('>I4sBBBI', 15, b'colr', 1, 2, 0, 276)
                ofile.write(buffer)

                # Write the rest of the boxes as-is.
                ifile.seek(77)
                ofile.write(ifile.read())
                ofile.flush()

        with self.assertWarns(UserWarning):
            Jp2k(ofile.name)

    def test_stupid_windows_eol_at_end(self):
        """
        SCENARIO:  An otherwise valid JP2 file has invalid bytes appended to
        the end of the file.  The number of bytes is less than 8 because any
        more than that would be interpreted as a box.

        SCENARIO:  A warning is issued.
        """
        with open(self.temp_jp2_filename, mode='wb') as ofile:
            with open(self.jp2file, 'rb') as ifile:
                # Copy the file all the way until the end.
                ofile.write(ifile.read())

                # then append a few extra bytes
                ofile.write(b'\0')
                ofile.flush()

            with self.assertWarns(UserWarning):
                Jp2k(ofile.name)

    @unittest.skipIf(OPENJPEG_NOT_AVAILABLE, OPENJPEG_NOT_AVAILABLE_MSG)
    def test_NR_ENC_X_6_2K_24_FULL_CBR_CIRCLE_000_tif_17_encode(self):
        """
        SCENARIO:  Too much data is written as a Cinema2K file.

        EXPECTED RESULT:  A warning from the openjpeg library is issued.
        """
        # Need to provide the proper size image
        data = glymur.Jp2k(self.jp2file)[:]
        data = np.concatenate((data, data), axis=0)
        data = np.concatenate((data, data), axis=1).astype(np.uint16)
        data = data[:1080, :2048, :]

        with self.assertWarns(UserWarning):
            Jp2k(self.temp_jp2_filename, data=data, cinema2k=24)

    def test_deprecated_set_get_printoptions(self):
        """
        Verify deprecated get_printoptions and set_printoptions
        """
        with self.assertWarns(DeprecationWarning):
            glymur.set_printoptions(short=True)
        with self.assertWarns(DeprecationWarning):
            glymur.set_printoptions(xml=True)
        with self.assertWarns(DeprecationWarning):
            glymur.set_printoptions(codestream=True)
        with self.assertWarns(DeprecationWarning):
            glymur.get_printoptions()

    def test_deprecated_set_get_parseoption(self):
        """
        Verify deprecated get_parseoptions and set_parseoptions
        """
        with self.assertWarns(DeprecationWarning):
            glymur.set_parseoptions(full_codestream=True)
        with self.assertWarns(DeprecationWarning):
            glymur.get_parseoptions()


class TestSuiteXML(unittest.TestCase):
    """
    This test should be run on both python2 and python3.
    """
    def test_bom(self):
        """
        Byte order markers are illegal in UTF-8.  Issue 185

        Original test file was input/nonregression/issue171.jp2
        """
        fptr = BytesIO()

        buffer = b"<?xpacket "
        buffer += b"begin='" + codecs.BOM_UTF8 + b"' "
        buffer += b"id='W5M0MpCehiHzreSzNTczkc9d'?>"
        buffer += b"<stuff>goes here</stuff>"
        buffer += b"<?xpacket end='w'?>"

        fptr.write(buffer)
        num_bytes = fptr.tell()
        fptr.seek(0)

        with warnings.catch_warnings(record=True) as w:
            glymur.jp2box.XMLBox.parse(fptr, 0, 8 + num_bytes)
            self.assertEqual(len(w), 0)



# standard library imports
import io
import logging
import pathlib
import struct
from typing import Tuple

# 3rd party library imports
import imageio.v3 as iio

# local imports
from .jp2k import Jp2k
from .options import set_option
from ._core_converter import _2JP2Converter


class JPEG2JP2(_2JP2Converter):
    """
    Attributes
    ----------
    create_exif_uuid : bool
        Create a UUIDBox for the Exif metadata.  Always True for JPEG.
    jp2_filename : path
        Path to JPEG 2000 file to be written.
    jpeg_filename : path
        Path to JPEG file.
    tilesize : tuple
        The dimensions of a tile in the JP2K file.
    verbosity : int
        Set the level of logging, i.e. WARNING, INFO, etc.
    tags : dict
        Tags retrieved from APP1 segment, if any.
    """
    def __init__(
        self,
        jpeg: pathlib.Path | str,
        jp2: pathlib.Path | str,
        include_icc_profile: bool = False,
        num_threads: int = 1,
        tilesize: Tuple[int, int] | None = None,
        verbosity: int = logging.CRITICAL,
        **kwargs
    ):
        super().__init__(True, True, include_icc_profile, tilesize, verbosity)

        self.jpeg_path = pathlib.Path(jpeg)

        self.jp2_path = pathlib.Path(jp2)
        if self.jp2_path.exists():
            raise FileExistsError(f'{str(self.jp2_path)} already exists, please delete if you wish to overwrite.')  # noqa : E501

        self.jp2_kwargs = kwargs

        self.tags = None

        # This is never set for JPEG
        self.exclude_tags = None

        if num_threads > 1:
            set_option("lib.num_threads", num_threads)

    def __enter__(self):
        """The JPEG2JP2 object must be used with a context manager."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass

    def run(self):

        self.copy_image()
        self.copy_metadata()

    def copy_metadata(self):
        """Transfer any EXIF or XMP metadata from the APPx segments."""

        with self.jpeg_path.open(mode='rb') as f:

            eof = False
            while not eof:

                marker = f.read(2)

                match marker:

                    case b'\xff\xd8':
                        # marker-only, SOI
                        pass

                    case b'\xff\xe0' | b'\xff\xec' | b'\xff\xee':
                        self.process_appx_segment(marker, f)

                    case b'\xff\xe1':
                        # EXIF using APP1
                        self.process_app1_segment(f)

                    case b'\xff\xe2':
                        # ICC profile
                        self.process_app2_segment(f)

                    case _:
                        # We don't care about anything else.  No need to scan
                        # the file any further, we're done.
                        eof = True

        if self.include_icc_profile and self.icc_profile is not None:
            self.rewrap_for_icc_profile()

    def process_appx_segment(self, marker, f):
        # APP0 (JFIF) is b'\xff\xe0'
        # APP12 ducky(?) is b'\xff\xec'
        # APP14 Adobe(?) is b'\xff\xee'
        _, n = struct.unpack('BB', marker)

        msg = f'Skipping APP{n - 224} segment...'
        self.logger.info(msg)

        data = f.read(2)
        size, = struct.unpack('>H', data)
        _ = f.read(size - 2)

    def process_app1_segment(self, f):
        """
        An APP1 segment can contain Exif or XMP data.
        """

        data = f.read(2)
        size, = struct.unpack('>H', data)
        buffer = f.read(size - 2)

        if buffer[:6] == b'Exif\x00\x00':

            # ok it is Exif

            buffer = buffer[6:]

            bf = io.BytesIO(buffer)

            self.read_tiff_header(bf)
            self.tags = self.read_ifd(bf)
            self.append_exif_uuid_box()

        elif buffer[:28] == b'http://ns.adobe.com/xap/1.0/':

            # XMP APP segment
            self.xmp_data = buffer[29:]
            self.append_xmp_uuid_box()

        else:

            offset = f.tell() - 2 - 2 - len(buffer)
            msg = f'Unrecognized APP1 segment at offset {offset}'
            self.logger.warning(msg)

    def process_app2_segment(self, f):
        """
        The APP2 segment(s) usually contains an ICC profile.  It may be split
        across more than one APP2 segment.
        """
        data = f.read(2)
        size, = struct.unpack('>H', data)
        buffer = f.read(size - 2)

        if buffer[:12] == b'ICC_PROFILE\x00':

            count, nchunks = struct.unpack('BB', buffer[12:14])

            if not self.include_icc_profile:
                msg = (
                    f'ICC profile chunk {count} of {nchunks} detected '
                    '(skipped)'
                )
                self.logger.warning(msg)
                return

            msg = f'Processing ICC profile chunk {count} of {nchunks}...'
            self.logger.info(msg)

            if count == 1:
                self.icc_profile = b''

            # accumulate the ICC profile stored in this chunk.  it's likely
            # that this is all that there is, though.
            self.icc_profile += bytes(buffer[14:])

    def copy_image(self):
        """Transfer the image data from the JPEG to the JP2 file."""
        image = iio.imread(self.jpeg_path)

        self.jp2 = Jp2k(
            self.jp2_path,
            tilesize=self.tilesize,
            **self.jp2_kwargs
        )

        self.jp2[:] = image



"""Manage glymur configuration settings."""

# Standard library imports
import copy
import warnings

# Local imports
from . import version
from .lib import openjp2 as opj2


_original_options = {
    "lib.num_threads": 1,
    "parse.full_codestream": False,
    "print.xml": True,
    "print.codestream": True,
    "print.short": False,
}
_options = copy.deepcopy(_original_options)


def set_option(key, value):
    """Set the value of the specified option.

    Available options:

        parse.full_codestream
        print.xml
        print.codestream
        print.short
        lib.num_threads

    Parameters
    ----------
    key : str
        Name of a single option.
    value :
        New value of option.

    Option Descriptions
    -------------------
    lib.num_threads : int
        Set the number of threads used to decode an image.  This option is only
        available with OpenJPEG 2.2.0 or higher.
    parse.full_codestream : bool
        When False, only the codestream header is parsed for metadata.  This
        can results in faster JP2/JPX parsing.  When True, the entire
        codestream is parsed. [default: False]
    print.codestream : bool
        When False, the codestream segments are not printed.  Otherwise the
        segments are printed depending on the value of the
        parse.full_codestream option. [default: True]
    print.short : bool
        When True, only the box ID, offset, and length are displayed.  Useful
        for displaying only the basic structure or skeleton of a JPEG 2000
        file. [default: False]
    print.xml : bool
        When False, printing of the XML contents of any XML boxes or UUID XMP
        boxes is suppressed. [default: True]

    Examples
    --------
    >>> glymur.set_option('print.short', True)
    >>> glymur.reset_option('all')

    See also
    --------
    get_option
    """
    if key not in _options.keys():
        raise KeyError(f"{key} not valid.")

    if key == "lib.num_threads":
        if version.openjpeg_version < "2.2.0":
            msg = (
                f"Thread support is not available on versions of OpenJPEG "
                f"prior to 2.2.0.  Your version is "
                f"{version.openjpeg_version}."
            )
            raise RuntimeError(msg)
        if not opj2.has_thread_support():
            msg = "The OpenJPEG library is not configured with thread support."
            raise RuntimeError(msg)

    _options[key] = value


def get_option(key):
    """Return the value of the specified option

    Available options:

        parse.full_codestream
        print.xml
        print.codestream
        print.short

    Parameter
    ---------
    key : str
        Name of a single option.

    Returns
    -------
    result : the value of the option.

    See also
    --------
    set_option
    """
    return _options[key]


def reset_option(key):
    """Reset one or more options to their default value.

    Pass "all" as argument to reset all options.

    Available options:

        parse.full_codestream
        print.xml
        print.codestream
        print.short

    Parameter
    ---------
    key : str
        Name of a single option.
    """
    global _options
    if key == "all":
        _options = copy.deepcopy(_original_options)
    else:
        if key not in _options.keys():
            raise KeyError(f"{key} not valid.")
        _options[key] = _original_options[key]


def set_parseoptions(full_codestream=True):
    warnings.warn(
        "Use set_option instead of set_parseoptions.", DeprecationWarning
    )
    set_option("parse.full_codestream", full_codestream)


def get_parseoptions():
    warnings.warn(
        "Use set_option instead of set_parseoptions.", DeprecationWarning
    )
    return {"full_codestream": get_option("parse.full_codestream")}


def set_printoptions(**kwargs):
    warnings.warn(
        "Use set_option instead of set_printoptions.", DeprecationWarning
    )
    for key, value in kwargs.items():
        if key not in ["short", "xml", "codestream"]:
            raise KeyError(f'"{key}" not a valid keyword parameter.')
        set_option("print." + key, value)


def get_printoptions():
    warnings.warn(
        "Use get_option instead of get_printoptions.", DeprecationWarning
    )
    d = {}
    for key in ["short", "xml", "codestream"]:
        d[key] = _options["print." + key]
    return d



"""
Module for tests specifically devoted to cinema profile.
"""

# Standard library imports
import unittest
import warnings

# 3rd party library imports
import numpy as np

# Local imports
import glymur
from glymur import Jp2k
from glymur.codestream import SIZsegment
from . import fixtures


class CinemaBase(fixtures.TestCommon):

    def verify_cinema_cod(self, cod_segment):

        self.assertFalse(cod_segment.scod & 2)  # no sop
        self.assertFalse(cod_segment.scod & 4)  # no eph
        self.assertEqual(cod_segment.prog_order, glymur.core.CPRL)
        self.assertEqual(cod_segment.layers, 1)
        self.assertEqual(cod_segment.mct, 1)
        self.assertEqual(cod_segment.num_res, 5)  # levels
        self.assertEqual(tuple(cod_segment.code_block_size), (32, 32))

    def _verify_siz_segment(self, actual, expected):
        """
        Verify the fields of the SIZ segment.
        """
        for field in [
            "rsiz",
            "xsiz",
            "ysiz",
            "xosiz",
            "yosiz",
            "xtsiz",
            "ytsiz",
            "xtosiz",
            "ytosiz",
            "bitdepth",
            "xrsiz",
            "yrsiz",
        ]:
            self.assertEqual(getattr(actual, field), getattr(expected, field))

    def check_cinema4k_codestream(self, codestream, image_size):

        kwargs = {
            "rsiz": 4,
            "xysiz": image_size,
            "xyosiz": (0, 0),
            "xytsiz": image_size,
            "xytosiz": (0, 0),
            "bitdepth": (12, 12, 12),
            "signed": (False, False, False),
            "xyrsiz": [(1, 1, 1), (1, 1, 1)],
        }
        self._verify_siz_segment(codestream.segment[1], SIZsegment(**kwargs))

        self.verify_cinema_cod(codestream.segment[2])

    def check_cinema2k_codestream(self, codestream, image_size):

        kwargs = {
            "rsiz": 3,
            "xysiz": image_size,
            "xyosiz": (0, 0),
            "xytsiz": image_size,
            "xytosiz": (0, 0),
            "bitdepth": (12, 12, 12),
            "signed": (False, False, False),
            "xyrsiz": [(1, 1, 1), (1, 1, 1)],
        }
        self._verify_siz_segment(codestream.segment[1], SIZsegment(**kwargs))

        self.verify_cinema_cod(codestream.segment[2])


@unittest.skipIf(
    fixtures.OPENJPEG_NOT_AVAILABLE, fixtures.OPENJPEG_NOT_AVAILABLE_MSG
)
class WriteCinema(CinemaBase):

    @classmethod
    def setUpClass(cls):
        cls.jp2_data = glymur.Jp2k(glymur.data.nemo())[:]

    def test_cinema2K_bad_frame_rate(self):
        """
        SCENARIO:  The cinema2k frame rate is not either 24 or 48.

        EXPECTED RESULT:  ValueError
        """
        with self.assertRaises(ValueError):
            Jp2k(self.temp_j2k_filename, data=self.jp2_data, cinema2k=36)

    def test_NR_ENC_X_6_2K_24_FULL_CBR_CIRCLE_000_tif_17_encode(self):
        """
        SCENARIO:  create JP2 file with cinema2k profile at 24 fps

        EXPECTED RESULT:  JP2 file has cinema2k profile

        The openjpeg test suite used the following input file for this test,
        input/nonregression/X_6_2K_24_FULL_CBR_CIRCLE_000.tif
        """
        # Need to provide the proper size image
        data = np.concatenate((self.jp2_data, self.jp2_data), axis=0)
        data = np.concatenate((data, data), axis=1).astype(np.uint16)
        data = data[:1080, :2048, :]

        with warnings.catch_warnings():
            # Ignore a warning issued by the library.
            warnings.simplefilter("ignore")
            j = Jp2k(self.temp_jp2_filename, data=data, cinema2k=24)

        codestream = j.get_codestream()
        self.check_cinema2k_codestream(codestream, (2048, 1080))

    def test_NR_ENC_X_6_2K_24_FULL_CBR_CIRCLE_000_tif_20_encode(self):
        """
        SCENARIO:  create JP2 file with cinema2k profile at 48 fps

        EXPECTED RESULT:  JP2 file has cinema2k profile

        The openjpeg test suite used the following input file for this test,
        input/nonregression/X_6_2K_24_FULL_CBR_CIRCLE_000.tif
        """
        # Need to provide the proper size image
        data = np.concatenate((self.jp2_data, self.jp2_data), axis=0)
        data = np.concatenate((data, data), axis=1).astype(np.uint16)
        data = data[:1080, :2048, :]

        with warnings.catch_warnings():
            # Ignore a warning issued by the library.
            warnings.simplefilter("ignore")
            j = Jp2k(self.temp_j2k_filename, data=data, cinema2k=48)

        codestream = j.get_codestream()
        self.check_cinema2k_codestream(codestream, (2048, 1080))

    def test_NR_ENC_ElephantDream_4K_tif_21_encode(self):
        """
        SCENARIO:  create JP2 file with cinema4k profile

        EXPECTED RESULT:  JP2 file has cinema4k profile

        The openjpeg test suite used the following input file for this test,
        input/nonregression/ElephantDream_4K.tif
        """
        # Need to provide the proper size image
        data = np.concatenate((self.jp2_data, self.jp2_data), axis=0)
        data = np.concatenate((data, data), axis=1).astype(np.uint16)
        data = data[:2160, :4096, :]

        with warnings.catch_warnings():
            # Ignore a warning issued by the library.
            warnings.simplefilter("ignore")
            j = Jp2k(self.temp_j2k_filename, data=data, cinema4k=True)

        codestream = j.get_codestream()
        self.check_cinema4k_codestream(codestream, (4096, 2160))
